﻿using System;
using UnityEngine;
using UnityEngine.Video;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MeshVR;
using SimpleJSON;

namespace uqone
{
    public class DisableCUAColliders : MVRScript
    {
        bool initialized = false;
        bool showDebugText = true; //false
        
        public override void Init()
        {
            CreateScriptCaptionUI(new JSONStorableString("scriptCaption",
                "<b><size=40>Disable CUA Colliders</size></b>\n" +
                "Enable and disable specific CUA colliders."));

            if (!showDebugText) CreateSpacer(true).height = 120f;

            StartCoroutine(WaitForAssetAndInit());
        }
        
        private UIDynamicTextField CreateScriptCaptionUI(JSONStorableString caption)
        {
            var result = CreateTextField(caption);
            result.height = 120f;
            result.textColor = new Color(255, 255, 255, 0.6f);
            //result.backgroundColor = new Color(0, 0, 0, 0.6f);
            result.backgroundColor = new Color(0.8f, 0, 0, 1.0f);

            return result;
        }

        List<Renderer> allRenderers = new List<Renderer>();

        protected IEnumerator WaitForAssetAndInit()
        {
            while (allRenderers.Count < 1)
            {// Wait until the asset is fully loaded before trying to add video...
                var sourceObject = containingAtom.containingSubScene == null ? containingAtom.gameObject : containingAtom.parentAtom.gameObject;
                sourceObject.GetComponentsInChildren<Renderer>(false, allRenderers);
                yield return new WaitForSeconds(0.5f);
            }
            
            // Wait until json load is done
            yield return new WaitForEndOfFrame();
            
            GenerateRendererOptionsList();
            
            initialized = true;
        }
        
        class RenderersOption
        {
            public string name;
            public string shortName;
            public bool isSkinned;
            public Mesh mesh;
            public GameObject go;
            public MeshRenderer meshRenderer;
            public SkinnedMeshRenderer skinnedMeshRenderer;
            public bool enabled;
            public JSONStorableBool enabledJSON;
        }
        
        List<RenderersOption> renderers = new List<RenderersOption>();
        
        /*private readonly List<string> MAKER_NAMES = new List<string>(new string[] {
            "FG ",
            "ST ",
            "DAS ",
            "TNB ",
            ""
        });*/
        
        void RemoveMakerName(string makerName)
        {
            foreach (RenderersOption r in renderers)
            {
                int pos = r.shortName.IndexOf(makerName);
                
                if (pos >= 0) {
                    r.shortName = r.shortName.Remove(0, makerName.Length);
                }
            }
        }
        
        void TryAutoRemoveUnknownMakerNames()
        {
            List<string> names = new List<string>();
            
            foreach (RenderersOption r in renderers)
            {
                // Find ones using a space separator...
                int pos = r.shortName.IndexOf(" ");
                
                if (pos >= 0) {
                    string shortName = r.shortName.Remove(pos) + " ";
                    
                    if (shortName == "Static ")
                    {// Ignore this one, static meshes are handled already...
                        continue;
                    }
                    
                    if (!names.Contains(shortName))
                    {
                        names.Add(shortName);
                    }
                }
                
                // Find ones using a underscore separator...
                pos = r.shortName.IndexOf("_");
                
                if (pos >= 0) {
                    string shortName = r.shortName.Remove(pos) + "_";
                    
                    if (!names.Contains(shortName))
                    {
                        names.Add(shortName);
                    }
                }
                
                // Find ones using a dash separator...
                pos = r.shortName.IndexOf("-");
                
                if (pos >= 0) {
                    string shortName = r.shortName.Remove(pos) + "-";
                    
                    if (!names.Contains(shortName))
                    {
                        names.Add(shortName);
                    }
                }
            }
            
            // Find if any are used in more than 75% of the mesh names, if they are, then we can be pretty sure we have a maker name...
            foreach (string name in names)
            {
                //SuperController.LogMessage("DisableCUAColliders: Maker name possibility: " + name);
                
                int numRepeats = 0;
                
                foreach (RenderersOption r in renderers)
                {
                    int pos = r.shortName.IndexOf(name);
                
                    if (pos >= 0) {
                        numRepeats++;
                    }
                }
                
                if (numRepeats >= renderers.Count() * 0.65f)
                {// This one is reused a lot, assume it is a maker name and remove it...
                    RemoveMakerName(name);
                }
                else if (name.Length <= 4 && numRepeats >= 8)
                {// This one is reused multiple times, but it is short, assume it is a maker name and remove it...
                    RemoveMakerName(name);
                }
            }
        }
        
        string MakeShortName(string name)
        {
            string shortName = name;
            
            //
            // Meshes merged using my plugin and other mesh merging systems...
            //
            
            // Remove "CombinedMeshes_" from the start of names...
            shortName = shortName.Replace("CombinedMeshes_", "");
            
            // Remove hashes from the end of names (merged meshes)...
            int pos = shortName.IndexOf("_-");  
            if (pos >= 0) {
                shortName = shortName.Remove(pos);  
            }
            
            //
            // Remove ".Shape"...
            //
            
            shortName = shortName.Replace(".Shape", "");
            
            //
            // Unity static meshes...
            //
            
            // Remove "Combined Mesh " from the start of names...
            shortName = shortName.Replace("Combined Mesh (root: scene) ", "Static Mesh ");
            shortName = shortName.Replace("Combined Mesh (root: scene)_", "Static Mesh ");
            shortName = shortName.Replace("Combined Mesh (root: scene)", "Static Mesh ");
            
            //
            // UltraScatter...
            //
            
            // Remove "Combined Mesh " from the start of names...
            shortName = shortName.Replace("UltraScatter [S] ", "");
            shortName = shortName.Replace("UltraScatter ", "");
            
            //
            // Clean up known asset creator names/tags from the start of names...
            //
            
            /*
            //
            // I now made an automated method of doing this, so we don't need this one...
            //
            
            // Clean up any make name/alias crap from the start of names...
            foreach (string makerName in MAKER_NAMES)
            {
                pos = shortName.IndexOf(makerName);  
                if (pos >= 0) {
                    shortName = shortName.Remove(0, makerName.Length);
                }
            }
            */
            
            return shortName;
        }
        
        void AddRendererOption(GameObject go, MeshRenderer meshRenderer, Mesh mesh, bool isSkinned, SkinnedMeshRenderer skinnedMeshRenderer)
        {
            try
            {
                RenderersOption newOption = new RenderersOption();
                newOption.name = mesh.name;
                
                if (newOption.name == null || newOption.name == "")
                {// An unnamed mesh, switch to using material name...
                    if (isSkinned)
                    {
                        if (skinnedMeshRenderer.sharedMaterials != null)
                        {
                            newOption.name = skinnedMeshRenderer.sharedMaterials[0].name;
                        }
                        
                        if ((newOption.name == null || newOption.name == "") && skinnedMeshRenderer.sharedMaterial != null)
                        {
                            newOption.name = skinnedMeshRenderer.sharedMaterial.name;
                        }
                        
                        if ((newOption.name == null || newOption.name == "") && skinnedMeshRenderer.material != null)
                        {
                            newOption.name = skinnedMeshRenderer.material.name;
                        }
                    }
                    else
                    {
                        if (meshRenderer.sharedMaterials != null)
                        {
                            newOption.name = meshRenderer.sharedMaterials[0].name;
                        }
                        
                        if ((newOption.name == null || newOption.name == "") && meshRenderer.sharedMaterial != null)
                        {
                            newOption.name = meshRenderer.sharedMaterial.name;
                        }
                        
                        if ((newOption.name == null || newOption.name == "") && meshRenderer.material != null)
                        {
                            newOption.name = meshRenderer.material.name;
                        }
                    }
                }
                
                //SuperController.LogMessage("DisableCUAColliders: Option " + newOption.name);
                
                int instanceNum = 0;
                string possibleName = "";
                
                // Clean up any "_instance" duplicates...
                int pos = newOption.name.IndexOf("_instance");  
                if (pos >= 0) {  
                    newOption.name = newOption.name.Remove(pos);  
                }
                
                // Clean up any "_duplicate" duplicates...
                pos = newOption.name.IndexOf("_duplicate");  
                if (pos >= 0) {  
                    newOption.name = newOption.name.Remove(pos);  
                }
                
                // Clean up any "_dup" duplicates...
                pos = newOption.name.IndexOf("_dup");  
                if (pos >= 0) {  
                    newOption.name = newOption.name.Remove(pos);  
                }
                
                while (true)
                {
                    possibleName = newOption.name + "_" + instanceNum;
                    
                    RenderersOption r = GetRendererOptionFromName(possibleName);
                    
                    if (r == null)
                    {
                        break;
                    }
                    
                    instanceNum++;
                }
                
                newOption.name = possibleName;
                newOption.shortName = MakeShortName(newOption.name);
                newOption.isSkinned = isSkinned;
                newOption.mesh = mesh;
                newOption.go = go;
                newOption.enabled = isSkinned ? skinnedMeshRenderer.enabled : meshRenderer.enabled;
                newOption.meshRenderer = meshRenderer;
                newOption.skinnedMeshRenderer = skinnedMeshRenderer;
                
                //SuperController.LogMessage("DisableCUAColliders: Final " + newOption.shortName);
                
                renderers.Add(newOption);
            }
            catch (Exception e) {
                SuperController.LogError("Init Exception caught: " + e);
            }
        }
        
        RenderersOption GetRendererOptionFromName(string name)
        {
            foreach (RenderersOption r in renderers)
            {
                if (r.name == name)
                {
                    return r;
                }
            }
            return null;
        }
        
        void CreateUI()
        {
            if (showDebugText)
            {
                JSONStorableString debugTextJSON = new JSONStorableString("", "<b> </b>\n" + "There are " + renderers.Count() + " renderers with colliders in this scene.");
                UIDynamicTextField debugText = CreateTextField(debugTextJSON, true);
                debugText.textColor = new Color(255, 255, 255, 0.6f);
                debugText.backgroundColor = new Color(0, 0, 0.8f, 1.0f);
                debugText.height = 120f;
            }
            
            //SuperController.LogMessage("DisableCUAColliders: There are " + renderers.Count() + " renderers in this scene.");
            
            bool right = false;
            
            foreach (RenderersOption r in renderers)
            {
                r.enabledJSON = new JSONStorableBool(r.shortName, r.enabled,  (b) => {
                    r.enabled = b;
                        
                    var colliders = r.go.GetComponentsInChildren(typeof(Collider));
                    
                    int numChanged = 0;
                    
                    foreach (Collider c in colliders)
                    {
                        if (c.enabled != b)
                        {
                            c.enabled = b;
                            numChanged++;
                        }
                    }
                        
                    //SuperController.LogMessage("DisableCUAColliders: " + numChanged + " colliders " + (b ? "ENABLED" : "DISABLED") + " on " + r.shortName + ".");
                });
                
                RegisterBool(r.enabledJSON);
                CreateToggle(r.enabledJSON, right);
                
                right = !right;
            }
        }
        
        void GenerateRendererOptionsList()
        {
            // Clear any old list...
            renderers.Clear();
            
            var sourceObject = containingAtom.containingSubScene == null ? containingAtom.gameObject : containingAtom.parentAtom.gameObject;
            GameObject GO = sourceObject.gameObject;
            
            //
            // Do all mesh filters/renderers...
            //
            var filters = GO.GetComponentsInChildren<MeshFilter>();
            
            //SuperController.LogMessage("Found " + filters.Length + " MeshFilters.");
            
            if (filters != null)
            {
                foreach (MeshFilter f in filters)
                {
                    GameObject obj = f.gameObject;
                    var renderer = obj.GetComponent<MeshRenderer>();
                    
                    if (renderer == null)
                    {
                        continue;
                    }
                    
                    Mesh mesh = f.sharedMesh;
                    
                    if (mesh == null)
                    {
                        if (f.mesh != null)
                        {
                            mesh = f.mesh;
                        }
                        else
                        {
                            continue;
                        }
                    }
                    
                    //var collider = obj.GetComponent<Collider>();
                    var colliders = obj.GetComponentsInChildren(typeof(Collider));
                    
                    if (colliders != null && colliders.Length > 0)
                    {
                        AddRendererOption(obj, renderer, mesh, false, null);
                    }
                }
            }
            
            
            //
            // Do all skinned meshes...
            //
            var skinnedRenderers = GO.GetComponentsInChildren<SkinnedMeshRenderer>();
            
            //SuperController.LogMessage("Found " + skinnedRenderers.Length + " SkinnedMeshRenderers.");
            
            if (skinnedRenderers != null)
            {
                foreach (SkinnedMeshRenderer r in skinnedRenderers)
                {
                    GameObject obj = r.gameObject;
                    
                    Mesh mesh = r.sharedMesh;

                    if (mesh == null)
                    {
                        continue;
                    }
                    
                    //var collider = obj.GetComponent<Collider>();
                    var colliders = obj.GetComponentsInChildren(typeof(Collider));
                    
                    if (colliders != null && colliders.Length > 0)
                    {
                        AddRendererOption(obj, null, mesh, true, r);
                    }
                }
            }
            
            // Try to remove maker names from the beginning of all the names...
            TryAutoRemoveUnknownMakerNames();
            
            // Now, sort the objects list by short names, for the UI...
            renderers.Sort((x, y) => x.shortName.CompareTo(y.shortName));
            
            CreateUI();
        }
        
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            
            if (includePhysical || forceStore)
            {
                jc["count"].AsInt = renderers.Count;
                
                //SuperController.LogMessage("Save count " + renderers.Count);
                
                int current = 0;
                
                foreach (RenderersOption r in renderers)
                {
                    jc["enabled" + current].AsBool = r.enabledJSON.val;
                    current++;
                }
                
                needsStore = true;
            }
            
            return jc;
        }
        
        protected IEnumerator LateRestoreFromJSONWait(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            while (!initialized)
            {// Wait until the asset is fully loaded before trying to add video...
                yield return new WaitForSeconds(0.5f);
            }
            
            // Wait until json load is done
            //yield return new WaitForEndOfFrame();
            
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);

            if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
            {
                int count = jc["count"].AsInt;
                
                //SuperController.LogMessage("Load count " + count);

                int current = 0;
                
                foreach (RenderersOption r in renderers)
                {// Load the trigger datas...
                    r.enabledJSON.val = jc["enabled" + current].AsBool;
                    
                    var colliders = r.go.GetComponentsInChildren(typeof(Collider));
                    
                    foreach (Collider c in colliders)
                        c.enabled = r.enabledJSON.val;
                        
                    current++;
                }
            }
        }
        
        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            StartCoroutine(LateRestoreFromJSONWait(jc, restorePhysical, restoreAppearance, setMissingToDefault));
        }
        
        /// <summary>
        /// Destroy created components on exit/reload
        /// </summary>
        void OnDestroy()
        {
            foreach (RenderersOption r in renderers)
            {// Re-enable them all...
                var colliders = r.go.GetComponentsInChildren(typeof(Collider));
                
                foreach (Collider c in colliders)
                    c.enabled = true;
            }
        }
        
        public void OnDisable()
        {
            OnDestroy();
        }
        
        public void OnEnable()
        {
            foreach (RenderersOption r in renderers)
            {// Load the trigger datas...
                var colliders = r.go.GetComponentsInChildren(typeof(Collider));
                
                foreach (Collider c in colliders)
                    c.enabled = r.enabledJSON.val;
            }
        }
        
        private void OnSceneLoaded()
        {
            StartCoroutine(WaitForAssetAndInit());
        }
    }
}
