using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using UnityEngine.Events;
using System;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;

namespace Unique
{
    public class DisableCUAIntercollisions : MVRScript
    {
        public string pluginAuthor = "unique";
        public string pluginName = "DisableCUAIntercollisions";
        public string pluginVersion = "1.0";
        public string pluginDate = "2022-01-30";
        public string pluginDescription =
 @"
 Disables inter-collisions between all custom unity assets.
 ";
        
        List<Collider> allColliders = new List<Collider>();
        
        int numActiveColliders = 0;
        public Collider[] activeColliders;
        
        private float myClock = -1.0f;
        
        // Use this for initialization
        public override void Init()
        {
            // Usage text...
            JSONStorableString usageInfoTextJSON = new JSONStorableString("", "\nDisables inter-collisions between all custom unity assets.\n\nDoes not stop other atom collisions with the objects, and skips any objects with rigidbodies.\n\nShould be added as either a Scene or Session plugin.\n\nNothing to configure here.");
            UIDynamicTextField usageInfoText = CreateTextField(usageInfoTextJSON);
            usageInfoText.backgroundColor = new Color32(60, 58, 210, 255);
            usageInfoText.height = 420f;
            usageInfoText.textColor = Color.white;
            
            Update();
        }

        void Update()
        {
            myClock -= Time.deltaTime;
            
            if (myClock > 0.0f)
            {// wait
                return;
            }
            
            myClock += 10.0f;
            myClock = Mathf.Max(myClock, 0.0f);
            
            DisableInterCollisions();
        }
        
        void CreateLists()
        {
            try
            {
                // Fist, clear out any old lists...
                OnDestroy();
                
                foreach (Atom atom in FindObjectsOfType(typeof(Atom)))
                {
                    //SuperController.LogMessage("atom = " + atom.name + " atomtype = " + atom.type);
                  
                    if (atom.type == "CustomUnityAsset")
                    {
                        foreach (Collider collider in atom.GetComponentsInChildren<BoxCollider>())
                        {
                            allColliders.Add(collider);
                        }
                        
                        foreach (Collider collider in atom.GetComponentsInChildren<CapsuleCollider>())
                        {
                            allColliders.Add(collider);
                        }
                        
                        foreach (Collider collider in atom.GetComponentsInChildren<SphereCollider>())
                        {
                            allColliders.Add(collider);
                        }

                        foreach (Collider collider in atom.GetComponentsInChildren<TerrainCollider>())
                        {
                            allColliders.Add(collider);
                        }
                        
                        foreach (Collider collider in atom.GetComponentsInChildren<MeshCollider>())
                        {
                            allColliders.Add(collider);
                        }
                    }
                }
                
                // We can't use List<> to pass to the unity function though, so we have to convert, add them to an active collider list...
                numActiveColliders = 0;
                activeColliders = new Collider[allColliders.Count];
                
                foreach (Collider activeCollider in allColliders)
                {
                    activeColliders[numActiveColliders] = activeCollider;
                    numActiveColliders++;
                }
                
                //SuperController.LogMessage("DisableCUAIntercollisions: Found " + numActiveColliders + " colliders for scene CUAs.");
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        void DisableInterCollisions()
        {
            try
            {
                CreateLists();
                
                if (numActiveColliders > 0)
                {
                    IgnoreCollision(activeColliders);
                    //SuperController.LogMessage("DisableInterCollisions: Disabled inter-collisions between " + numActiveColliders + " CUA colliders.");
                }
                /*else
                {
                    SuperController.LogMessage("DisableInterCollisions: No active CUA colliders.");
                }*/
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        
        void EnableInterCollisions()
        {
            try
            {
                if (numActiveColliders > 0)
                {
                    AllowCollision(activeColliders);
                    //SuperController.LogMessage("EnableInterCollisions: Re-enabled inter-collisions between " + numActiveColliders + " CUA colliders.");
                    allColliders.Clear();
                    numActiveColliders = 0;
                }
                /*else
                {
                    SuperController.LogMessage("EnableInterCollisions: No inactive CUA colliders.");
                }*/
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        
        void OnDestroy()
        {
            EnableInterCollisions();
        }
        
        #region IgnoreCollision
        public static void IgnoreCollision(params Collider[] colliders)
        {
           IgnoreCollision(true, colliders);
        }

        public static void AllowCollision(params Collider[] colliders)
        {
           IgnoreCollision(false, colliders);
        }

        public static void IgnoreCollision(bool ignore, params Collider[] colliders)
        {
           for(int i = 0; i < colliders.Length - 1; i++)
           {
               for(int j = i + 1; j < colliders.Length; j++)
               {
                   Physics.IgnoreCollision(colliders[i], colliders[j], ignore);
               }
           }
        }
        #endregion
    }
}
