﻿using System;
using UnityEngine;
using UnityEngine.Video;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MeshVR;
using SimpleJSON;

namespace uqone
{
    public class DisableCUAObjects2 : MVRScript
    {
        bool initialized = false;
        bool showDebugText = true; //false
        
        UIDynamicButton destroyButton;
        public JSONStorableAction actionDestroy;
        
        UIDynamicButton uncheckButton;
        public JSONStorableAction actionUncheck;
        
        public override void Init()
        {
            CreateScriptCaptionUI(new JSONStorableString("scriptCaption",
                "<b><size=40>Disable CUA Objects 2.1</size></b>\n" +
                "Enable/Disable (or destroy) CUA mesh components and colliders."));

            if (!showDebugText) CreateSpacer(true).height = 120f;
            
            StartCoroutine(WaitForAssetAndInit());
        }
        
        private UIDynamicTextField CreateScriptCaptionUI(JSONStorableString caption)
        {
            var result = CreateTextField(caption);
            result.height = 120f;
            //result.textColor = new Color(255, 255, 255, 0.6f);
            //result.backgroundColor = new Color(0, 0, 0, 0.6f);
            
            result.textColor = new Color(255, 255, 255, 1.0f);
            result.backgroundColor = new Color(0, 0, 255, 1.0f);

            return result;
        }

        List<Renderer> allRenderers = new List<Renderer>();

        protected IEnumerator WaitForAssetAndInit()
        {
            while (allRenderers.Count < 1)
            {// Wait until the asset is fully loaded before trying to add video...
                var sourceObject = containingAtom.containingSubScene == null ? containingAtom.gameObject : containingAtom.parentAtom.gameObject;
                sourceObject.GetComponentsInChildren<Renderer>(false, allRenderers);
                yield return new WaitForSeconds(0.5f);
            }
            
            // Wait until json load is done
            yield return new WaitForEndOfFrame();
            
            GenerateRendererOptionsList();
            
            initialized = true;
        }
        
        class RenderersOption
        {
            public string name;
            public string shortName;
            public bool isSkinned;
            public Mesh mesh;
            public GameObject go;
            public MeshRenderer meshRenderer;
            public SkinnedMeshRenderer skinnedMeshRenderer;
            public bool enabled;
            public JSONStorableBool enabledJSON;
            public UIDynamicToggle uiToggle;
        }
        
        List<RenderersOption> renderers = new List<RenderersOption>();
        
        /*private readonly List<string> MAKER_NAMES = new List<string>(new string[] {
            "FG ",
            "ST ",
            "DAS ",
            "TNB ",
            ""
        });*/
        
        void RemoveMakerName(string makerName)
        {
            foreach (RenderersOption r in renderers)
            {
                int pos = r.shortName.IndexOf(makerName);
                
                if (pos >= 0) {
                    r.shortName = r.shortName.Remove(0, makerName.Length);
                }
            }
        }
        
        void TryAutoRemoveUnknownMakerNames()
        {
            List<string> names = new List<string>();
            
            foreach (RenderersOption r in renderers)
            {
                // Find ones using a space separator...
                int pos = r.shortName.IndexOf(" ");
                
                if (pos >= 0) {
                    string shortName = r.shortName.Remove(pos) + " ";
                    
                    if (shortName == "Static ")
                    {// Ignore this one, static meshes are handled already...
                        continue;
                    }
                    
                    if (!names.Contains(shortName))
                    {
                        names.Add(shortName);
                    }
                }
                
                // Find ones using a underscore separator...
                pos = r.shortName.IndexOf("_");
                
                if (pos >= 0) {
                    string shortName = r.shortName.Remove(pos) + "_";
                    
                    if (!names.Contains(shortName))
                    {
                        names.Add(shortName);
                    }
                }
                
                // Find ones using a dash separator...
                pos = r.shortName.IndexOf("-");
                
                if (pos >= 0) {
                    string shortName = r.shortName.Remove(pos) + "-";
                    
                    if (!names.Contains(shortName))
                    {
                        names.Add(shortName);
                    }
                }
            }
            
            // Find if any are used in more than 75% of the mesh names, if they are, then we can be pretty sure we have a maker name...
            foreach (string name in names)
            {
                //SuperController.LogMessage("DisableCUAObjects2: Maker name possibility: " + name);
                
                int numRepeats = 0;
                
                foreach (RenderersOption r in renderers)
                {
                    int pos = r.shortName.IndexOf(name);
                
                    if (pos >= 0) {
                        numRepeats++;
                    }
                }
                
                if (numRepeats >= renderers.Count() * 0.65f)
                {// This one is reused a lot, assume it is a maker name and remove it...
                    RemoveMakerName(name);
                }
                else if (name.Length <= 4 && numRepeats >= 8)
                {// This one is reused multiple times, but it is short, assume it is a maker name and remove it...
                    RemoveMakerName(name);
                }
            }
        }
        
        string MakeShortName(string name)
        {
            string shortName = name;
            
            //
            // Meshes merged using my plugin and other mesh merging systems...
            //
            
            // Remove "CombinedMeshes_" from the start of names...
            shortName = shortName.Replace("CombinedMeshes_", "");
            
            // Remove hashes from the end of names (merged meshes)...
            int pos = shortName.IndexOf("_-");  
            if (pos >= 0) {
                shortName = shortName.Remove(pos);  
            }
            
            //
            // Remove ".Shape"...
            //
            
            shortName = shortName.Replace(".Shape", "");
            
            //
            // Unity static meshes...
            //
            
            // Remove "Combined Mesh " from the start of names...
            shortName = shortName.Replace("Combined Mesh (root: scene) ", "Static Mesh ");
            shortName = shortName.Replace("Combined Mesh (root: scene)_", "Static Mesh ");
            shortName = shortName.Replace("Combined Mesh (root: scene)", "Static Mesh ");
            
            //
            // UltraScatter...
            //
            
            // Remove "Combined Mesh " from the start of names...
            shortName = shortName.Replace("UltraScatter [S] ", "");
            shortName = shortName.Replace("UltraScatter ", "");
            
            //
            // Clean up known asset creator names/tags from the start of names...
            //
            
            /*
            //
            // I now made an automated method of doing this, so we don't need this one...
            //
            
            // Clean up any make name/alias crap from the start of names...
            foreach (string makerName in MAKER_NAMES)
            {
                pos = shortName.IndexOf(makerName);  
                if (pos >= 0) {
                    shortName = shortName.Remove(0, makerName.Length);
                }
            }
            */
            
            return shortName;
        }
        
        void AddRendererOption(GameObject go, MeshRenderer meshRenderer, Mesh mesh, bool isSkinned, SkinnedMeshRenderer skinnedMeshRenderer)
        {
            try
            {
                RenderersOption newOption = new RenderersOption();
                newOption.name = mesh.name;
                
                if (newOption.name == null || newOption.name == "")
                {// An unnamed mesh, switch to using material name...
                    if (isSkinned)
                    {
                        if (skinnedMeshRenderer.sharedMaterials != null)
                        {
                            newOption.name = skinnedMeshRenderer.sharedMaterials[0].name;
                        }
                        
                        if ((newOption.name == null || newOption.name == "") && skinnedMeshRenderer.sharedMaterial != null)
                        {
                            newOption.name = skinnedMeshRenderer.sharedMaterial.name;
                        }
                        
                        if ((newOption.name == null || newOption.name == "") && skinnedMeshRenderer.material != null)
                        {
                            newOption.name = skinnedMeshRenderer.material.name;
                        }
                    }
                    else
                    {
                        if (meshRenderer.sharedMaterials != null)
                        {
                            newOption.name = meshRenderer.sharedMaterials[0].name;
                        }
                        
                        if ((newOption.name == null || newOption.name == "") && meshRenderer.sharedMaterial != null)
                        {
                            newOption.name = meshRenderer.sharedMaterial.name;
                        }
                        
                        if ((newOption.name == null || newOption.name == "") && meshRenderer.material != null)
                        {
                            newOption.name = meshRenderer.material.name;
                        }
                    }
                }
                
                //SuperController.LogMessage("DisableCUAObjects2: Option " + newOption.name);
                
                int instanceNum = 0;
                string possibleName = "";
                
                // Clean up any "_instance" duplicates...
                int pos = newOption.name.IndexOf("_instance");  
                if (pos >= 0) {  
                    newOption.name = newOption.name.Remove(pos);  
                }
                
                // Clean up any "_duplicate" duplicates...
                pos = newOption.name.IndexOf("_duplicate");  
                if (pos >= 0) {  
                    newOption.name = newOption.name.Remove(pos);  
                }
                
                // Clean up any "_dup" duplicates...
                pos = newOption.name.IndexOf("_dup");  
                if (pos >= 0) {  
                    newOption.name = newOption.name.Remove(pos);  
                }
                
                while (true)
                {
                    possibleName = newOption.name + "_" + instanceNum;
                    
                    RenderersOption r = GetRendererOptionFromName(possibleName);
                    
                    if (r == null)
                    {
                        break;
                    }
                    
                    instanceNum++;
                }
                
                newOption.name = possibleName;
                newOption.shortName = MakeShortName(newOption.name);
                newOption.isSkinned = isSkinned;
                newOption.mesh = mesh;
                newOption.go = go;
                newOption.enabled = isSkinned ? skinnedMeshRenderer.enabled : meshRenderer.enabled;
                newOption.meshRenderer = meshRenderer;
                newOption.skinnedMeshRenderer = skinnedMeshRenderer;
                newOption.uiToggle = null;
                
                //SuperController.LogMessage("DisableCUAObjects2: Final " + newOption.shortName);
                
                renderers.Add(newOption);
            }
            catch (Exception e) {
                SuperController.LogError("Init Exception caught: " + e);
            }
        }
        
        RenderersOption GetRendererOptionFromName(string name)
        {
            foreach (RenderersOption r in renderers)
            {
                if (r.name == name)
                {
                    return r;
                }
            }
            return null;
        }
        
        void CreateUI()
        {
            if (showDebugText)
            {
                JSONStorableString debugTextJSON = new JSONStorableString("", "<b><size=40> </size></b>\n" + "There are " + renderers.Count() + " renderers in this CUA.");
                UIDynamicTextField debugText = CreateTextField(debugTextJSON, true);
                debugText.height = 120f;
                debugText.backgroundColor = new Color32(0, 96, 0, 255);
                debugText.textColor = new Color32(255, 255, 255, 255);
            }
            
            
            destroyButton = CreateButton("Destroy Disabled Objects", false);
			destroyButton.button.onClick.AddListener(DestroyDisabled);
			destroyButton.buttonColor = new Color32(255, 0, 0, 255);
			destroyButton.textColor = new Color32(255, 255, 255, 255);
                
			actionDestroy = new JSONStorableAction("Destroy Disabled Objects", () => { DestroyDisabled(); });
			RegisterAction(actionDestroy);
			
			//CreateSpacer(true).height = 52f;
			
			uncheckButton = CreateButton("Deselect All Objects", true);
			uncheckButton.button.onClick.AddListener(DisableAll);
			uncheckButton.buttonColor = new Color32(0, 0, 255, 255);
			uncheckButton.textColor = new Color32(255, 255, 255, 255);
                
			actionUncheck = new JSONStorableAction("Deselect All Objects", () => { DisableAll(); });
			RegisterAction(actionUncheck);
			
            
            //SuperController.LogMessage("DisableCUAObjects2: There are " + renderers.Count() + " renderers in this scene.");
            
            bool right = false;
            
            foreach (RenderersOption r in renderers)
            {
                r.enabledJSON = new JSONStorableBool(r.shortName, r.enabled,  (b) => {
                    r.enabled = b;
                    
                    if (r.isSkinned)
                        r.skinnedMeshRenderer.enabled = b;
                    else
                        r.meshRenderer.enabled = b;
                        
                    var colliders = r.go.GetComponents<Collider>();
                    
                    foreach (Collider c in colliders)
                        c.enabled = b;
                        
                    //SuperController.LogMessage("DisableCUAObjects2: " + colliders.Count() + " colliders " + (b ? "ENABLED" : "DISABLED"));
                });
                
                RegisterBool(r.enabledJSON);
                r.uiToggle = CreateToggle(r.enabledJSON, right);
                
                right = !right;
            }
        }
        
        void GenerateRendererOptionsList()
        {
            // Clear any old list...
            renderers.Clear();
            
            var sourceObject = containingAtom.containingSubScene == null ? containingAtom.gameObject : containingAtom.parentAtom.gameObject;
            GameObject GO = sourceObject.gameObject;
            
            //
            // Do all mesh filters/renderers...
            //
            var filters = GO.GetComponentsInChildren<MeshFilter>();
            
            //SuperController.LogMessage("Found " + filters.Length + " MeshFilters.");
            
            if (filters != null)
            {
                foreach (MeshFilter f in filters)
                {
                    GameObject obj = f.gameObject;
                    var renderer = obj.GetComponent<MeshRenderer>();
                    
                    if (renderer == null)
                    {
                        continue;
                    }
                    
                    Mesh mesh = f.sharedMesh;
                    
                    if (mesh == null)
                    {
                        if (f.mesh != null)
                        {
                            mesh = f.mesh;
                        }
                        else
                        {
                            continue;
                        }
                    }
                    
                    AddRendererOption(obj, renderer, mesh, false, null);
                }
            }
            
            
            //
            // Do all skinned meshes...
            //
            var skinnedRenderers = GO.GetComponentsInChildren<SkinnedMeshRenderer>();
            
            //SuperController.LogMessage("Found " + skinnedRenderers.Length + " SkinnedMeshRenderers.");
            
            if (skinnedRenderers != null)
            {
                foreach (SkinnedMeshRenderer r in skinnedRenderers)
                {
                    GameObject obj = r.gameObject;
                    
                    Mesh mesh = r.sharedMesh;

                    if (mesh == null)
                    {
                        continue;
                    }
                    
                    AddRendererOption(obj, null, mesh, true, r);
                }
            }
            
            // Try to remove maker names from the beginning of all the names...
            TryAutoRemoveUnknownMakerNames();
            
            // Now, sort the objects list by short names, for the UI...
            renderers.Sort((x, y) => x.shortName.CompareTo(y.shortName));
            
            CreateUI();
        }
        
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            
            if (includePhysical || forceStore)
            {
                jc["count"].AsInt = renderers.Count;
                
                //SuperController.LogMessage("Save count " + renderers.Count);
                
                int current = 0;
                
                foreach (RenderersOption r in renderers)
                {
                    jc["enabled" + current].AsBool = r.enabledJSON.val;
                    current++;
                }
                
                needsStore = true;
            }
            
            return jc;
        }
        
        protected IEnumerator LateRestoreFromJSONWait(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            while (!initialized)
            {// Wait until the asset is fully loaded before trying to add video...
                yield return new WaitForSeconds(0.5f);
            }
            
            // Wait until json load is done
            //yield return new WaitForEndOfFrame();
            
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);

            if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
            {
                int count = jc["count"].AsInt;
                
                //SuperController.LogMessage("Load count " + count);

                int current = 0;
                
                foreach (RenderersOption r in renderers)
                {// Load the trigger datas...
                    r.enabledJSON.val = jc["enabled" + current].AsBool;
                    
                    if (r.isSkinned)
                        r.skinnedMeshRenderer.enabled = r.enabledJSON.val;
                    else
                        r.meshRenderer.enabled = r.enabledJSON.val;
                        
                    var colliders = r.go.GetComponents<Collider>();
                    
                    foreach (Collider c in colliders)
                        c.enabled = r.enabledJSON.val;
                        
                    current++;
                }
            }
        }
        
        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            StartCoroutine(LateRestoreFromJSONWait(jc, restorePhysical, restoreAppearance, setMissingToDefault));
        }
        
        void DestroyDisabled()
        {
			List<RenderersOption> destroyedRenderers = new List<RenderersOption>();
			int numRemovedMeshRenderers = 0;
			int numRemovedSkinnedRenderers = 0;
			int numRemovedColliders = 0;
			int numRemovedRigidbodies = 0;
			int numRemovedJoints = 0;
			int numRemovedAudios = 0;
			int numRemovedLights = 0;
			
			foreach (RenderersOption r in renderers)
            {// Destroy anything that is disabled...
                if (r.isSkinned && !r.enabledJSON.val && r.skinnedMeshRenderer != null)
                {
                    DestroyImmediate(r.skinnedMeshRenderer);
                    numRemovedSkinnedRenderers++;
					
					// Unity, you fucking suck, WHY ARE THESE NOT BEING DESTROYED?!?!?!?!?!? Piece of shit software...
					var js = r.go.GetComponents<Joint>();
                
					foreach (Joint j in js)
					{
						DestroyImmediate(j);
						numRemovedJoints++;
					}
					
					// Unity, you fucking suck, WHY ARE THESE NOT BEING DESTROYED?!?!?!?!?!? Piece of shit software...
					var rbs = r.go.GetComponents<Rigidbody>();
                
					foreach (Rigidbody rb in rbs)
					{
						DestroyImmediate(rb);
						numRemovedRigidbodies++;
					}
					
					var colliders = r.go.GetComponentsInChildren<Collider>();
                
					foreach (Collider c in colliders)
					{
						DestroyImmediate(c);
						numRemovedColliders++;
					}
					
					var ads = r.go.GetComponentsInChildren<AudioSource>();
                
					foreach (AudioSource a in ads)
					{
						DestroyImmediate(a);
						numRemovedAudios++;
					}
					
					var ls = r.go.GetComponentsInChildren<Light>();
                
					foreach (Light l in ls)
					{
						DestroyImmediate(l);
						numRemovedLights++;
					}

                    destroyedRenderers.Add(r);
				}
                else if (!r.enabledJSON.val && r.meshRenderer != null)
                {
                    DestroyImmediate(r.meshRenderer);
                    numRemovedMeshRenderers++;
                    
                    MeshFilter filter = r.go.GetComponent<MeshFilter>();
                    
                    if (filter != null)
                    {
						DestroyImmediate(filter);
                    }
                    
                    // Unity, you fucking suck, WHY ARE THESE NOT BEING DESTROYED?!?!?!?!?!? Piece of shit software...
					var js = r.go.GetComponents<Joint>();
                
					foreach (Joint j in js)
					{
						DestroyImmediate(j);
						numRemovedJoints++;
					}
					
					// Unity, you fucking suck, WHY ARE THESE NOT BEING DESTROYED?!?!?!?!?!? Piece of shit software...
					var rbs = r.go.GetComponents<Rigidbody>();
                
					foreach (Rigidbody rb in rbs)
					{
						DestroyImmediate(rb);
						numRemovedRigidbodies++;
					}
					
					Destroy(r.go.GetComponent<Rigidbody>());
					
					var colliders = r.go.GetComponentsInChildren<Collider>();
                
					foreach (Collider c in colliders)
					{
						DestroyImmediate(c);
						numRemovedColliders++;
					}
					
					var ads = r.go.GetComponentsInChildren<AudioSource>();
                
					foreach (AudioSource a in ads)
					{
						DestroyImmediate(a);
						numRemovedAudios++;
					}
					
					var ls = r.go.GetComponentsInChildren<Light>();
                
					foreach (Light l in ls)
					{
						DestroyImmediate(l);
						numRemovedLights++;
					}
						
                    destroyedRenderers.Add(r);
				}
            }
            
            // Now destroy the renderer options...
            foreach (RenderersOption r in destroyedRenderers)
            {// Destroy anything that is disabled...
				//r.uiToggle.disabled = true;
				//r.uiToggle.textColor = new Color(255, 255, 255, 1.0f);
				r.uiToggle.backgroundColor = new Color(128, 0, 0, 1.0f);
				r.uiToggle.toggle.interactable = false;
			}
            
            SuperController.LogMessage("DisableCUAObjects: Destroyed " + numRemovedMeshRenderers + " mesh renderers, " + numRemovedSkinnedRenderers + " skinned mesh renderers, " + numRemovedColliders + " colliders, " + numRemovedJoints + " joints, " + numRemovedRigidbodies + " rigid bodies, " + numRemovedAudios + " audio sources, and " + numRemovedLights + " lights from " + destroyedRenderers.Count + " disabled objects.");
            
            int removedGameobjectCount = 0;
            
            foreach (RenderersOption r in renderers)
            {//
				if (r.go != null)
				{
					removedGameobjectCount += ClearInactives(r.go);
				}
            }
            
            SuperController.LogMessage("Destroyed " + removedGameobjectCount + " inactive game objects from hierarchy.");
        }
        
        void KillTheFuckingThing(GameObject obj)
        {
			Transform TempParent = new GameObject().transform;
            obj.transform.SetParent(TempParent);
            DestroyImmediate(TempParent.gameObject);
        }
        
        public int ClearInactives(GameObject obj)
        {
            int removedCount = 0;
            
            if (obj == null)
            {
				return 0;
            }
            
            if (!obj.activeInHierarchy)
            {
                //SuperController.LogMessage("Removing inactive game object: \"" + obj.name + "\".");
                DestroyImmediate(obj);
                removedCount++;
                return removedCount;
            }
            
            var subObjects = obj.GetComponentsInChildren(typeof(Component));
            
            //SuperController.LogMessage("game object: " + obj.name + " has " + (subObjects != null ? subObjects.Count().ToString() : "0") + " subcomponents.");
            
            bool found = false;
            
            if (subObjects != null && subObjects.Count() > 2)
            {
				foreach (var o in subObjects)
				{
					//SuperController.LogMessage("   >>> " + o.GetType().ToString());
					
					if (o != null 
						&& o.GetType() != typeof(Transform) 
						&& o.GetType() != typeof(GameObject) 
						&& o.GetType() != typeof(Joint) 
						&& o.GetType() != typeof(AudioSource) 
						&& o.GetType() != typeof(Light) 
						&& o.GetType() != typeof(Rigidbody))
					{
						found = true;
					}
				}
            }
            
            if (!found)
            {
                //SuperController.LogMessage("Removing inactive game object: \"" + obj.name + "\".");
                GameObject.DestroyImmediate(obj);
                //GameObject.Destroy(obj, 0.0f);
                //KillTheFuckingThing(obj);
                removedCount++;
            }

            return removedCount;
        }
        
        void DisableAll()
        {
			foreach (RenderersOption r in renderers)
            {// Re-enable them all...
                if (r.isSkinned && r.skinnedMeshRenderer != null)
                    r.skinnedMeshRenderer.enabled = false;
                else if (r.meshRenderer != null)
                    r.meshRenderer.enabled = false;
                    
				if (r.go != null)
				{
					var colliders = r.go.GetComponents<Collider>();
					
					foreach (Collider c in colliders)
					{
						if (c != null)
						{
							c.enabled = false;
						}
					}
				}
				
				r.enabledJSON.val = false;
            }
        }
        
        /// <summary>
        /// Destroy created components on exit/reload
        /// </summary>
        void OnDestroy()
        {
            foreach (RenderersOption r in renderers)
            {// Re-enable them all...
                if (r.isSkinned && r.skinnedMeshRenderer != null)
                    r.skinnedMeshRenderer.enabled = true;
                else if (r.meshRenderer != null)
                    r.meshRenderer.enabled = true;
                    
				if (r.go != null)
				{
					var colliders = r.go.GetComponents<Collider>();
					
					foreach (Collider c in colliders)
					{
						if (c != null)
						{
							c.enabled = true;
						}
					}
				}
            }
        }
        
        public void OnDisable()
        {
            OnDestroy();
        }
        
        public void OnEnable()
        {
            foreach (RenderersOption r in renderers)
            {// Load the trigger datas...
                if (r.isSkinned && r.skinnedMeshRenderer != null)
                    r.skinnedMeshRenderer.enabled = r.enabledJSON.val;
                else if (r.meshRenderer != null)
                    r.meshRenderer.enabled = r.enabledJSON.val;
                    
				if (r.go != null)
				{
					var colliders = r.go.GetComponents<Collider>();
					
					foreach (Collider c in colliders)
					{
						if (c != null)
						{
							c.enabled = r.enabledJSON.val;
						}
					}
				}
            }
        }
        
        private void OnSceneLoaded()
        {
            StartCoroutine(WaitForAssetAndInit());
        }
    }
}
