using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using UnityEngine.Events;
using System;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;

namespace Unique
{

    public class DisableCUAProbes : MVRScript
    {
        public string pluginAuthor = "unique";
        public string pluginName = "DisableCUAProbes";
        public string pluginVersion = "1.0";
        public string pluginDate = "2022-01-20";
        public string pluginDescription =
 @"
 Disable ReflectionProbes in CUA assets.
 ";
 
        UIDynamicButton updateButton;
        public JSONStorableAction actionUpdate;
        public JSONStorableBool destroyMethodJSON;
        
        // Use this for initialization
        public override void Init()
        {
            try
            {
                // Usage text...
                JSONStorableString usageInfoTextJSON = new JSONStorableString("", "\nDisables Reflection Probes in CUA.\n\nJust add this plugin to the CUA atom.\n\nIf destroy is selected re-enabling the probes will require reloading the asset, but it possibly also reduces overhead.");
                UIDynamicTextField usageInfoText = CreateTextField(usageInfoTextJSON);
                usageInfoText.backgroundColor = new Color32(60, 58, 210, 255);
                usageInfoText.height = 290f;
                usageInfoText.textColor = Color.white;
                
                updateButton = CreateButton("Update", false);
                updateButton.button.onClick.AddListener(DisableReflectionProbes);
                
                SuperController.singleton.onSceneLoadedHandlers += OnSceneLoaded;
                
                actionUpdate = new JSONStorableAction("Update", () => { DisableReflectionProbes(); });
                RegisterAction(actionUpdate);
                
                destroyMethodJSON = new JSONStorableBool("Destroy instead of disable?", false);
                destroyMethodJSON.storeType = JSONStorableParam.StoreType.Full;
                CreateToggle(destroyMethodJSON, true);
                RegisterBool(destroyMethodJSON);
                
                DisableReflectionProbes();
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        
        void OnDestroy()
        {
            var sourceObject = containingAtom.containingSubScene == null ? containingAtom.gameObject : containingAtom.parentAtom.gameObject;
            
            foreach (ReflectionProbe l in sourceObject.GetComponentsInChildren<ReflectionProbe>())
            {
                l.enabled = true;
            }
        }
        
        private void OnSceneLoaded()
        {
            DisableReflectionProbes();
        }
        
        public void OnEnable()
        {
            DisableReflectionProbes();
        }

        public void OnDisable()
        {
            OnDestroy();
        }

        void DisableReflectionProbes()
        {
            var sourceObject = containingAtom.containingSubScene == null ? containingAtom.gameObject : containingAtom.parentAtom.gameObject;
            
            int destroyed = 0;
            
            foreach (ReflectionProbe l in sourceObject.GetComponentsInChildren<ReflectionProbe>())
            {
                destroyed++;
                
                if (destroyMethodJSON.val)
                {
                    Destroy(l);
                }
                else
                {
                    l.enabled = false;
                }
            }
            
            if (destroyed > 0)
            {
                SuperController.LogMessage("DisableCUAProbes: Disabled/distroyed " + destroyed + " reflection probes.");
            }
        }
    }
}
