using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using UnityEngine.Events;
using System;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;

namespace Unique
{
    public class GenerateColliders : MVRScript
    {
        //JSONStorableBool replaceOldCollidersJSON;
        //JSONStorableBool replaceNonMeshCollidersJSON;
        JSONStorableBool transparentCollidersEnabledJSON;
        JSONStorableBool allowCutoutJSON;
        JSONStorableBool allowGlassJSON;
        JSONStorableBool heightRangeEnabledJSON;
        JSONStorableFloat minHeightJSON;
        JSONStorableFloat maxHeightJSON;
        JSONStorableBool convexCollidersEnabledJSON;
        UIDynamicButton resetButton;
        protected JSONStorableAction resetTrigger;
        
        /*enum BlendModes
        {// This causes all of Unity to crash without error, ffs... What a fkin joke.
           Opaque,
           Cutout,
           Fade,
           Transparent
        }*/
        
        // Use this for initialization
        public override void Init()
        {
            // Usage text...
            JSONStorableString usageInfoTextJSON = new JSONStorableString("", "\nGenerate missing colliders for any custom unity asset.");
            UIDynamicTextField usageInfoText = CreateTextField(usageInfoTextJSON);
            usageInfoText.backgroundColor = new Color32(60, 58, 210, 255);
            usageInfoText.height = 130f;
            usageInfoText.textColor = Color.white;
            
            /*replaceOldCollidersJSON = new JSONStorableBool("Replace Old Colliders", true);
            RegisterBool(replaceOldCollidersJSON);
            CreateToggle(replaceOldCollidersJSON, true);
            
            replaceNonMeshCollidersJSON = new JSONStorableBool("Replace Non-Mesh Colliders", false);
            RegisterBool(replaceNonMeshCollidersJSON);
            CreateToggle(replaceNonMeshCollidersJSON, true);*/
            
            JSONStorableString transTextJSON = new JSONStorableString("", "\nBy default transparent materials will be skipped. This can be overridden with these options.");
            UIDynamicTextField transText = CreateTextField(transTextJSON, true);
            transText.backgroundColor = new Color32(210, 58, 60, 255);
            transText.height = 155f;
            transText.textColor = Color.white;
            
            transparentCollidersEnabledJSON = new JSONStorableBool("Transparent Colliders", false);
            RegisterBool(transparentCollidersEnabledJSON);
            CreateToggle(transparentCollidersEnabledJSON, true);
            
            allowCutoutJSON = new JSONStorableBool("CutOut Colliders", false);
            RegisterBool(allowCutoutJSON);
            CreateToggle(allowCutoutJSON, true);
            
            allowGlassJSON = new JSONStorableBool("Names Containing \"Glass\"", true);
            RegisterBool(allowGlassJSON);
            CreateToggle(allowGlassJSON, true);
            
            JSONStorableString heightTextJSON = new JSONStorableString("", "\nBy default we will only generate colliders at reachable levels. This can be overridden with these options.");
            UIDynamicTextField heightText = CreateTextField(heightTextJSON, true);
            heightText.backgroundColor = new Color32(210, 58, 60, 255);
            heightText.height = 155f;
            heightText.textColor = Color.white;
            
            heightRangeEnabledJSON = new JSONStorableBool("Use Height Range", true);
            RegisterBool(heightRangeEnabledJSON);
            CreateToggle(heightRangeEnabledJSON, true);
            
            minHeightJSON = new JSONStorableFloat("Min Collision Height", -100.0f, -100.0f, 100.0f, true);
            minHeightJSON.storeType = JSONStorableParam.StoreType.Full;
            RegisterFloat(minHeightJSON);
            CreateSlider(minHeightJSON, true);
            
            maxHeightJSON = new JSONStorableFloat("Max Collision Height", 3.0f, -100.0f, 100.0f, true);
            maxHeightJSON.storeType = JSONStorableParam.StoreType.Full;
            RegisterFloat(maxHeightJSON);
            CreateSlider(maxHeightJSON, true);
            
            JSONStorableString convexTextJSON = new JSONStorableString("", "\nConvex colliders reduce the CPU usage on collisions at the cost of accuracy. Great for smaller objects.\n\nNot recommended for interiors, or environments.");
            UIDynamicTextField convexText = CreateTextField(convexTextJSON, true);
            convexText.backgroundColor = new Color32(210, 58, 60, 255);
            convexText.height = 265f;
            convexText.textColor = Color.white;
            
            convexCollidersEnabledJSON = new JSONStorableBool("Use Convex Colliders", false);
            RegisterBool(convexCollidersEnabledJSON);
            CreateToggle(convexCollidersEnabledJSON, true);
            
            JSONStorableString resetTextJSON = new JSONStorableString("", "\nUse the reset button to delete and regenerate all colliders after changing options.");
            UIDynamicTextField resetText = CreateTextField(resetTextJSON, false);
            resetText.backgroundColor = new Color32(210, 58, 60, 255);
            resetText.height = 155f;
            resetText.textColor = Color.white;
            
            resetButton = CreateButton("Reset", false);
            resetButton.button.onClick.AddListener(ResetButtonCallback);
            
            // Make the reset button triggerable from other places...
            resetTrigger = new JSONStorableAction("Reset", () => { ResetButtonCallback(); });
            RegisterAction(resetTrigger);
            
            //GenerateAllColliders();
            
            // In case there is no "asset bundle loaded" callback yet, we have to wait
            StartCoroutine(WaitForAssetBundle());
        }
        
        List<Renderer> allRenderers = new List<Renderer>();
        
        protected IEnumerator WaitForAssetBundle()
        {
			allRenderers.Clear();
			
            while (allRenderers.Count < 1)
            {// Wait until the asset is fully loaded before trying to add video...
                try
                {
                    if (containingAtom != null)
                    {
                        var sourceObject = containingAtom.containingSubScene == null ? containingAtom.gameObject : containingAtom.parentAtom.gameObject;
                        sourceObject.GetComponentsInChildren<Renderer>(false, allRenderers);
                    }
                }
                catch (Exception e)
                {
                    //SuperController.LogError($"{this.name} : Unhandled exception: " + e);
                }
                
                yield return new WaitForSeconds(0.5f);
            }
            
            OnDestroy();
            GenerateAllColliders();
        }

        void ResetButtonCallback()
        {
            // In case there is no "asset bundle loaded" callback yet, we have to wait
            StartCoroutine(WaitForAssetBundle());
        }
                
        void GenerateAllColliders()
        {
            try
            {
                // Some sanity checking...
                if (minHeightJSON.val > maxHeightJSON.val)
                {
                    minHeightJSON.val = maxHeightJSON.val - 1.0f;
                }
                
                if (maxHeightJSON.val < minHeightJSON.val)
                {
                    maxHeightJSON.val = minHeightJSON.val + 1.0f;
                }
                
                var sourceObject = containingAtom.containingSubScene == null ? containingAtom.gameObject : containingAtom.parentAtom.gameObject;
                GameObject GO = sourceObject.gameObject;
                
                //
                // Do all mesh filter/renderer colliders...
                //
                var filters = GO.GetComponentsInChildren<MeshFilter>();
                
                //SuperController.LogMessage("Found " + filters.Length + " MeshFilters.");
                
                if (filters != null)
                {
                    for (int i = 0; i < filters.Length; i++)
                    {
                        GameObject obj = filters[i].gameObject;
                        //SuperController.LogMessage("Found " + obj.name);
                        var renderer = obj.GetComponent<Renderer>();
                        
                        if (filters[i].sharedMesh == null || renderer == null)
                        {
                            return;
                        }
                        
                        GenerateCollider(obj, filters[i].sharedMesh, renderer.sharedMaterials);
                    }
                }
                
                //
                // Do all skinned mesh colliders...
                //
                var skinnedRenderers = GO.GetComponentsInChildren<SkinnedMeshRenderer>();
                
                //SuperController.LogMessage("Found " + skinnedRenderers.Length + " SkinnedMeshRenderers.");
                
                if (skinnedRenderers != null)
                {
                    for (int i = 0; i < skinnedRenderers.Length; i++)
                    {
                        GameObject obj = skinnedRenderers[i].gameObject;
                        //SuperController.LogMessage("Found " + obj.name);

                        if (skinnedRenderers[i].sharedMesh == null || skinnedRenderers[i].sharedMaterials == null)
                        {
                            return;
                        }
                        
                        GenerateCollider(obj, skinnedRenderers[i].sharedMesh, skinnedRenderers[i].sharedMaterials);
                    }
                }
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        
        bool MaterialValidForCollider(Material material)
        {
            if (!transparentCollidersEnabledJSON.val)
            {// Transparent material collisions are disabled...
                var blendMode = (int)Math.Round(material.GetFloat("_Mode")); // WTF is this a FLOAT, Unity!?!?!?!?!!?
                //var blendMode = material.GetInt("_Mode");
                
                if (allowCutoutJSON.val && blendMode == 1 /*Cutout*/)
                {// Allow cutout...
                    
                }
                else if (allowGlassJSON.val && material.name.IndexOf("glass", StringComparison.OrdinalIgnoreCase) >= 0)
                {// Allow glass...
                
                }
                /*else if (material.renderQueue == (int)UnityEngine.Rendering.RenderQueue.Transparent)
                {// Skip transparent objects...
                    return false;
                }*/
                else if (blendMode >= 2 /* Fade/Trans*/ || blendMode == 1 /*Cutout*/)
                {// Skip transparent objects...
                    return false;
                }
                
            }

            return true;
        }
        
        public void GenerateCollider(GameObject GO, Mesh sharedMesh, Material[] sharedMaterials)
        {
            try
            {
                if (sharedMesh == null)
                {
                    return;
                }
                
                var totalsubmeshcount = sharedMaterials.Length;
                
                if (totalsubmeshcount <= 0)
                {// Nothing to do here...
                    return;
                }
                
                // move required submeshes lower
                var materials = sharedMaterials;
                
                // Check if there are any valid surfaces here...
                bool foundValidMateiral = false;
                
                for (int i = 0; i < totalsubmeshcount; i++)
                {
                    if (MaterialValidForCollider(materials[i]))
                    {
                        foundValidMateiral = true;
                        break;
                    }
                }
                
                if (!foundValidMateiral)
                {// Nothing to do here...
                    return;
                }
                
                Quaternion OldRot = GO.transform.rotation;
                Vector3 OldPos = GO.transform.position;

                GO.transform.rotation = Quaternion.identity;
                GO.transform.position = Vector3.zero;

                var originalMeshName = sharedMesh.name;

                // Make the new mesh...
                var newMesh = new Mesh();
                //newMesh.name = string.Format("GeneratedCollider_{0}_{1}", originalMeshName, (int)Mathf.Abs(newMesh.GetHashCode()));
                //newMesh.name = string.Format("GeneratedCollider_{0}", (int)Mathf.Abs(newMesh.GetHashCode()));
                newMesh.name = string.Format("GeneratedCollider_{0}", originalMeshName);
                newMesh.indexFormat = UnityEngine.Rendering.IndexFormat.UInt32;
                
                List<int> newTris = new List<int>();
                List<Vector3> newVerts = new List<Vector3>();
                List<Vector3> newNormals = new List<Vector3>();

                int current_vert = 0;

                for (int i = 0; i < totalsubmeshcount; i++)
                {
                    if (!MaterialValidForCollider(materials[i]))
                    {// Skip this one...
                        continue;
                    }
                    
                    var oldVerts = new List<Vector3>();
                    sharedMesh.GetVertices(oldVerts);

                    var oldTris = sharedMesh.GetTriangles(i);

                    var oldNormals = new List<Vector3>();
                    sharedMesh.GetNormals(oldNormals);

                    for (int j = 0; j < oldTris.Length; j+=3)
                    {
                        var oldTri1 = oldTris[j];
                        var oldVert1 = oldVerts[oldTri1];
                        var oldNormal1 = oldNormals[oldTri1];

                        var oldTri2 = oldTris[j+1];
                        var oldVert2 = oldVerts[oldTri2];
                        var oldNormal2 = oldNormals[oldTri2];

                        var oldTri3 = oldTris[j + 2];
                        var oldVert3 = oldVerts[oldTri3];
                        var oldNormal3 = oldNormals[oldTri3];

                        if (heightRangeEnabledJSON.val && oldVert1.y < minHeightJSON.val && oldVert2.y < minHeightJSON.val && oldVert3.y < minHeightJSON.val)
                        {
                            continue;
                        }
                        
                        if (heightRangeEnabledJSON.val && oldVert1.y > maxHeightJSON.val && oldVert2.y > maxHeightJSON.val && oldVert3.y > maxHeightJSON.val)
                        {
                            continue;
                        }

                        newVerts.Add(oldVert1);
                        newNormals.Add(oldNormal1);
                        newTris.Add(current_vert);
                        current_vert++;

                        newVerts.Add(oldVert2);
                        newNormals.Add(oldNormal2);
                        newTris.Add(current_vert);
                        current_vert++;

                        newVerts.Add(oldVert3);
                        newNormals.Add(oldNormal3);
                        newTris.Add(current_vert);
                        current_vert++;
                    }
                }
                
                if (current_vert <= 0 || newVerts.Count <= 0 || newTris.Count <= 0)
                {// Seems there was nothing to add (max height most likely)...
                    newMesh = null;
                    
                    GO.transform.rotation = OldRot;
                    GO.transform.position = OldPos;
                    return;
                }
                else
                {
                    newMesh.SetVertices(newVerts);
                    newMesh.SetTriangles(newTris, 0);
                    newMesh.SetNormals(newNormals);

                    var collider = GO.AddComponent<MeshCollider>();
                    collider.sharedMesh = null;
                    collider.sharedMesh = newMesh;
                    
                    if (convexCollidersEnabledJSON.val)
                    {// See if we can make a convex collider...
                        try
                        {// Inside the try in case it fails, it should fall back...
                            collider.inflateMesh = true;
                            collider.skinWidth = 0.0001f;
                            collider.convex = true;
                            
                            SuperController.LogMessage("GenerateCollider: Created collider mesh " + newMesh.name + " with " + newVerts.Count + " verts and " + newTris.Count + " tris.");
                            SuperController.LogMessage("GenerateCollider: A convex collider was generated successfully from the above mesh.");
                        }
                        catch (Exception e)
                        {
                            collider.inflateMesh = false;
                            collider.convex = false;
                            
                            SuperController.LogMessage("GenerateCollider: Created collider mesh " + newMesh.name + " with " + newVerts.Count + " verts and " + newTris.Count + " tris.");
                            SuperController.LogMessage("GenerateCollider: Convex collider failed to generate for this mesh, using full mesh for collisions.");
                        }
                    }
                    else
                    {
                        SuperController.LogMessage("GenerateCollider: Created collider mesh " + newMesh.name + " with " + newVerts.Count + " verts and " + newTris.Count + " tris.");
                    }
                    
                    GO.transform.rotation = OldRot;
                    GO.transform.position = OldPos;
                }
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        
        void OnDestroy()
        {
            try
            {
                if (containingAtom == null)
                {
                    return;
                }
                
                var sourceObject = containingAtom.containingSubScene == null ? containingAtom.gameObject : containingAtom.parentAtom.gameObject;
                
                if (sourceObject == null)
                {
                    return;
                }
                
                GameObject GO = sourceObject.gameObject;
                
                if (GO != null)
                {
                    //if (replaceOldCollidersJSON.val && replaceNonMeshCollidersJSON.val)
                    {
                        var oldBoxColls = GO.GetComponentsInChildren<BoxCollider>();
                        
                        foreach (BoxCollider collider in oldBoxColls)
                        {
                            if (collider != null)
                            {
                                Destroy(collider);
                            }
                        }
                        
                        var oldCapsuleColls = GO.GetComponentsInChildren<CapsuleCollider>();
                        
                        foreach (CapsuleCollider collider in oldCapsuleColls)
                        {
                            if (collider != null)
                            {
                                Destroy(collider);
                            }
                        }
                        
                        var oldSphereColls = GO.GetComponentsInChildren<SphereCollider>();
                        
                        foreach (SphereCollider collider in oldSphereColls)
                        {
                            if (collider != null)
                            {
                                Destroy(collider);
                            }
                        }
                        
                        var oldTerrainColls = GO.GetComponentsInChildren<TerrainCollider>();
                        
                        foreach (TerrainCollider collider in oldTerrainColls)
                        {
                            if (collider != null)
                            {
                                Destroy(collider);
                            }
                        }
                    }
                    
                    var colliders = GO.GetComponentsInChildren<MeshCollider>();
                    
                    if (colliders != null && colliders.Length > 0)
                    {
                        foreach (MeshCollider collider in colliders)
                        {
                            if (collider != null)
                            {
                                //if (replaceOldCollidersJSON.val)
                                {
                                    SuperController.LogMessage("deleting collider: " + collider.sharedMesh.name);
                                    Destroy(collider);
                                }
                                /*else if (collider.sharedMesh != null)
                                {
                                    if (collider.sharedMesh.name != null)
                                    {
                                        if (collider.sharedMesh.name.StartsWith("GeneratedCollider", StringComparison.OrdinalIgnoreCase))
                                        {// Only destroy stuff we made...
                                            SuperController.LogMessage("deleting collider: " + collider.sharedMesh.name);
                                            Destroy(collider);
                                        }
                                    }
                                }*/
                            }
                        }
                    }
                }
            }
            catch (Exception e)
            {
                //SuperController.LogError("Exception caught: " + e);
            }
        }
        
        public void OnEnable()
        {
            ResetButtonCallback();
        }

        public void OnDisable()
        {
            OnDestroy();
        }
        
        private void OnSceneLoaded()
        {
            ResetButtonCallback();
        }
    }
}
