using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using UnityEngine.Events;
using UnityEngine.Rendering;
using System;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using SimpleJSON;
using System.Linq;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;

namespace Unique
{
    public class OptimizeCUA : MVRScript
    {
        JSONStorableBool transparentCollidersEnabledJSON;
        JSONStorableBool allowCutoutJSON;
        JSONStorableBool allowGlassJSON;
        JSONStorableBool heightRangeEnabledJSON;
        JSONStorableFloat minHeightJSON;
        JSONStorableFloat maxHeightJSON;
        JSONStorableBool convexCollidersAllowedJSON;
        UIDynamicButton OptimizeCollidersButton;
        protected JSONStorableAction OptimizeCollidersJSON;
        
        UIDynamicButton CalculateNormalsButton;
        protected JSONStorableAction CalculateNormalsJSON;
        
        UIDynamicButton OptimizeButton;
        protected JSONStorableAction OptimizeJSON;

        // Use this for initialization
        public override void Init()
        {
            // Usage text...
            JSONStorableString titleTextJSON = new JSONStorableString("", "\nOptimize the number of draw calls for nearly any custom unity asset.");
            UIDynamicTextField titleText = CreateTextField(titleTextJSON);
            titleText.backgroundColor = new Color32(60, 58, 210, 255);
            titleText.height = 130f;
            titleText.textColor = Color.white;
            
            JSONStorableString usageInfoTextJSON = new JSONStorableString("", "\nBy merging all the meshes into a single mesh per material, we can optimize the render speed of the asset.\n\nOnce optimized, we then generate optimized colliders for the new meshes, ignoring all unreachable objects by default.");
            UIDynamicTextField usageInfoText = CreateTextField(usageInfoTextJSON);
            usageInfoText.backgroundColor = new Color32(60, 120, 58, 255);
            usageInfoText.height = 320f;
            usageInfoText.textColor = Color.white;
            
            JSONStorableString transTextJSON = new JSONStorableString("", "\nBy default transparent materials will be skipped. This can be overridden with these options.");
            UIDynamicTextField transText = CreateTextField(transTextJSON, true);
            transText.backgroundColor = new Color32(210, 58, 60, 255);
            transText.height = 155f;
            transText.textColor = Color.white;
            
            transparentCollidersEnabledJSON = new JSONStorableBool("Transparent Colliders", false);
            RegisterBool(transparentCollidersEnabledJSON);
            CreateToggle(transparentCollidersEnabledJSON, true);
            
            allowCutoutJSON = new JSONStorableBool("CutOut Colliders", false);
            RegisterBool(allowCutoutJSON);
            CreateToggle(allowCutoutJSON, true);
            
            allowGlassJSON = new JSONStorableBool("Names Containing \"Glass\"", true);
            RegisterBool(allowGlassJSON);
            CreateToggle(allowGlassJSON, true);
            
            JSONStorableString heightTextJSON = new JSONStorableString("", "\nBy default we will only generate colliders at reachable levels. This can be overridden with these options.");
            UIDynamicTextField heightText = CreateTextField(heightTextJSON, true);
            heightText.backgroundColor = new Color32(210, 58, 60, 255);
            heightText.height = 155f;
            heightText.textColor = Color.white;
            
            heightRangeEnabledJSON = new JSONStorableBool("Use Height Range", true);
            RegisterBool(heightRangeEnabledJSON);
            CreateToggle(heightRangeEnabledJSON, true);
            
            minHeightJSON = new JSONStorableFloat("Min Collision Height", -100.0f, -100.0f, 100.0f, true);
            minHeightJSON.storeType = JSONStorableParam.StoreType.Full;
            RegisterFloat(minHeightJSON);
            CreateSlider(minHeightJSON, true);
            
            maxHeightJSON = new JSONStorableFloat("Max Collision Height", 3.0f, -100.0f, 100.0f, true);
            maxHeightJSON.storeType = JSONStorableParam.StoreType.Full;
            RegisterFloat(maxHeightJSON);
            CreateSlider(maxHeightJSON, true);
            
            JSONStorableString convexTextJSON = new JSONStorableString("", "\nConvex colliders reduce the CPU usage on collisions at the cost of accuracy. Great for smaller objects.\n\nNot recommended for interiors, or environments.");
            UIDynamicTextField convexText = CreateTextField(convexTextJSON, true);
            convexText.backgroundColor = new Color32(210, 58, 60, 255);
            convexText.height = 265f;
            convexText.textColor = Color.white;
            
            convexCollidersAllowedJSON = new JSONStorableBool("Allow Convex Colliders", true);
            RegisterBool(convexCollidersAllowedJSON);
            CreateToggle(convexCollidersAllowedJSON, true);
            
            JSONStorableString resetTextJSON = new JSONStorableString("", "\nUse the Regenerate Colliders button to delete and regenerate all colliders after changing options.");
            UIDynamicTextField resetText = CreateTextField(resetTextJSON, false);
            resetText.backgroundColor = new Color32(210, 58, 60, 255);
            resetText.height = 155f;
            resetText.textColor = Color.white;
            
            OptimizeCollidersButton = CreateButton("Regenerate Colliders", false);
            OptimizeCollidersButton.button.onClick.AddListener(OptimizeCollidersCallback);
            
            // Make the Regenerate Colliders button triggerable from other places...
            OptimizeCollidersJSON = new JSONStorableAction("Regenerate Colliders", () => { OptimizeCollidersCallback(); });
            RegisterAction(OptimizeCollidersJSON);
            
            CalculateNormalsButton = CreateButton("Calculate Missing Normals", false);
            CalculateNormalsButton.button.onClick.AddListener(CalculateNormalsCallback);
            
            // Make the Regenerate Colliders button triggerable from other places...
            CalculateNormalsJSON = new JSONStorableAction("Calculate Missing Normals", () => { CalculateNormalsCallback(); });
            RegisterAction(CalculateNormalsJSON);
            
            OptimizeButton = CreateButton("Re-Optimize CUA", false);
            OptimizeButton.button.onClick.AddListener(OnEnable);
            
            // Make the Regenerate Colliders button triggerable from other places...
            OptimizeJSON = new JSONStorableAction("Re-Optimize CUA", () => { OnEnable(); });
            RegisterAction(OptimizeJSON);
            
            // In case there is no "asset bundle loaded" callback yet, we have to wait
            StartCoroutine(WaitForAssetBundle());
        }
        
        List<Renderer> allRenderers = new List<Renderer>();
        
        protected IEnumerator WaitForAssetBundle()
        {
            var sourceObject = containingAtom.containingSubScene == null ? containingAtom.gameObject : containingAtom.parentAtom.gameObject;
            
            while (allRenderers.Count < 1)
            {
                sourceObject.GetComponentsInChildren<Renderer>(false, allRenderers);
                yield return new WaitForSeconds(0.5f);
            }
            
            GameObject GO = sourceObject.gameObject;
            
            var filters = GO.GetComponentsInChildren<MeshFilter>();
            int numOriginalDrawCalls = 0;
            
            foreach (MeshFilter filter in filters)
            {
                var r = filter.gameObject.GetComponent<MeshRenderer>();
                var materials = r.sharedMaterials;
                numOriginalDrawCalls += materials.Length;
            }
            
            foreach (MeshFilter filter in filters)
            {
                //SuperController.LogMessage("DEBUG: Game object for filter " + filter.sharedMesh.name + " is " + filter.transform.gameObject.name);
                
                var parent = filter.transform.parent.gameObject;
                
                // Find the top-most gameobject where VAM stores the CUA mesh...
                while (parent && parent.gameObject.name != "rescaleObject")
                {
                    parent = parent.transform.parent.gameObject;
                }
                
                GO = parent;
                break;
            }
            
            // First clean out any empty game objects...
            CleanupEmptyGameObjects(GO);
            
            // Check for any combined material meshes, and split them as required, so they can be optimized...
            //SuperController.LogMessage("FixMissingMaterials");
            FixMissingMaterials(GO);
            
            if (!CheckMeshAccessability(GO))
            {
                SuperController.LogMessage("OptimizeCUA: Reading of some meshes was denied by the creator. Please slap them for their lack of openness. This may not work.");
            }
            
            // First, destroy all old colliders...
            //SuperController.LogMessage("DestroyColliders");
            DestroyColliders(GO);
        
            //SuperController.LogMessage("SplitMeshesIfRequired");
            SplitMeshesIfRequired(GO);
            
            //SuperController.LogMessage("CleanupDisabledRenderers");
            CleanupDisabledRenderers(GO);
            
            // Combine all meshes with the same materials to minimize draw calls...
            //SuperController.LogMessage("CombineMeshes");
            CombineMeshes(GO);
            
            //SuperController.LogMessage("CleanupDisabledRenderers");
            CleanupDisabledRenderers(GO);
            
            // Clean out any empty game objects, after our work is done...
            CleanupEmptyGameObjects(GO);
            
            //SuperController.LogMessage("CheckRecalculateNormals");
            CheckRecalculateNormals(GO);
            
            var filters2 = GO.GetComponentsInChildren<MeshFilter>();
            int numOptimizedDrawCalls = 0;
            
            foreach (MeshFilter filter in filters2)
            {
                var r = filter.gameObject.GetComponent<MeshRenderer>();
                var materials = r.sharedMaterials;
                numOptimizedDrawCalls += materials.Length;
            }
            
            SuperController.LogMessage("OptimizeCUA: Optimized " + numOriginalDrawCalls + " original CUA draw calls into " + numOptimizedDrawCalls + " optimized draw calls.");
            
            // Generate new colliders...
            //SuperController.LogMessage("OptimizeColliders");
            OptimizeColliders(GO);
            
            SuperController.LogMessage("OptimizeCUA: New colliders generated OK...");
        }
        
        void CalculateNormalsCallback()
        {
            var sourceObject = containingAtom.containingSubScene == null ? containingAtom.gameObject : containingAtom.parentAtom.gameObject;
            
            GameObject GO = sourceObject.gameObject;
            
            var filters = GO.GetComponentsInChildren<MeshFilter>();
            
            foreach (MeshFilter filter in filters)
            {
                //SuperController.LogMessage("DEBUG: Game object for filter " + filter.sharedMesh.name + " is " + filter.transform.gameObject.name);
                
                var parent = filter.transform.parent.gameObject;
                
                // Find the top-most gameobject where VAM stores the CUA mesh...
                while (parent && parent.gameObject.name != "rescaleObject")
                {
                    parent = parent.transform.parent.gameObject;
                }
                
                GO = parent;
                break;
            }
            
            if (!CheckMeshAccessability(GO))
            {
                SuperController.LogMessage("OptimizeCUA: Reading of meshes was denied by the creator. Please slap them for their lack of community openness. This may not work.");
            }
            
            CleanupDisabledRenderers(GO);
            
            CheckRecalculateNormals(GO);
        }

        void OptimizeCollidersCallback()
        {
            var sourceObject = containingAtom.containingSubScene == null ? containingAtom.gameObject : containingAtom.parentAtom.gameObject;
            
            GameObject GO = sourceObject.gameObject;
            
            var filters = GO.GetComponentsInChildren<MeshFilter>();
            
            foreach (MeshFilter filter in filters)
            {
                //SuperController.LogMessage("DEBUG: Game object for filter " + filter.sharedMesh.name + " is " + filter.transform.gameObject.name);
                
                var parent = filter.transform.parent.gameObject;
                
                // Find the top-most gameobject where VAM stores the CUA mesh...
                while (parent && parent.gameObject.name != "rescaleObject")
                {
                    parent = parent.transform.parent.gameObject;
                }
                
                GO = parent;
                break;
            }
            
            if (!CheckMeshAccessability(GO))
            {
                SuperController.LogMessage("OptimizeCUA: Reading of meshes was denied by the creator. Please slap them for their lack of community openness. This may not work.");
            }
            
            // First, destroy all old colliders...
            DestroyColliders(GO);
            
            CleanupDisabledRenderers(GO);

            // Generate new colliders...
            OptimizeColliders(GO);
            
            SuperController.LogMessage("OptimizeCUA: New colliders generated OK...");
        }

        void OptimizeColliders(GameObject sourceObject)
        {
            // Some sanity checking...
            if (minHeightJSON.val > maxHeightJSON.val)
            {
                minHeightJSON.val = maxHeightJSON.val - 1.0f;
            }
            
            if (maxHeightJSON.val < minHeightJSON.val)
            {
                maxHeightJSON.val = minHeightJSON.val + 1.0f;
            }
            
            //
            // Do all mesh filter/renderer colliders...
            //
            var filters = sourceObject.GetComponentsInChildren<MeshFilter>();
            
            //SuperController.LogMessage("Found " + filters.Length + " MeshFilters.");
            
            if (filters != null)
            {
                foreach (MeshFilter filter in filters)
                {
                    if (!filter.sharedMesh || !filter.sharedMesh.isReadable)
                    {
                        continue;
                    }
                    
                    GameObject obj = filter.gameObject;
                    //SuperController.LogMessage("Found " + obj.name);
                    
                    MeshRenderer renderer = obj.GetComponent<MeshRenderer>();
                    
                    if (filter.sharedMesh == null || renderer == null)
                    {
                        //return;
                        continue;
                    }
                    
                    GenerateCollider(obj, filter.sharedMesh, renderer.sharedMaterials);
                }
            }
            
            //
            // Do all skinned mesh colliders...
            //
            /*var skinnedRenderers = sourceObject.GetComponentsInChildren<SkinnedMeshRenderer>();
            
            //SuperController.LogMessage("Found " + skinnedRenderers.Length + " SkinnedMeshRenderers.");
            
            if (skinnedRenderers != null)
            {
                foreach (SkinnedMeshRenderer smr in skinnedRenderers)
                {
                    GameObject obj = smr.gameObject;
                    //SuperController.LogMessage("Found " + obj.name);

                    if (smr.sharedMesh == null || smr.sharedMaterials == null)
                    {
                        return;
                    }
                    
                    GenerateCollider(obj, smr.sharedMesh, smr.sharedMaterials);
                }
            }*/
        }
        
        bool MaterialValidForCollider(Material material)
        {
            if (material == null)
            {
                return false;
            }
            
            if (!transparentCollidersEnabledJSON.val)
            {// Transparent material collisions are disabled...
                var blendMode = (int)Math.Round(material.GetFloat("_Mode")); // WTF is this a FLOAT, Unity!?!?!?!?!!?
                //var blendMode = material.GetInt("_Mode");
                
                if (allowCutoutJSON.val && blendMode == 1 /*Cutout*/)
                {// Allow cutout...
                    
                }
                else if (allowGlassJSON.val && material.name.IndexOf("glass", StringComparison.OrdinalIgnoreCase) >= 0)
                {// Allow glass...
                
                }
                /*else if (material.renderQueue == (int)UnityEngine.Rendering.RenderQueue.Transparent)
                {// Skip transparent objects...
                    return false;
                }*/
                else if (blendMode >= 2 /* Fade/Trans*/ || blendMode == 1 /*Cutout*/)
                {// Skip transparent objects...
                    return false;
                }
                
            }

            return true;
        }
        
        bool GenerateCollider(GameObject GO, Mesh sharedMesh, Material[] sharedMaterials)
        {
            try
            {
                if (sharedMesh == null)
                {
                    return false;
                }
                
                var totalsubmeshcount = sharedMaterials.Length;
                
                if (totalsubmeshcount <= 0)
                {// Nothing to do here...
                    return false;
                }
                
                // move required submeshes lower
                var materials = sharedMaterials;
                
                if (materials == null || materials.Length < 1)
                {
                    SuperController.LogMessage(sharedMesh.name + " has no materials.");
                    return false;
                }
                
                // Check if there are any valid surfaces here...
                bool foundValidMateiral = false;
                
                for (int i = 0; i < totalsubmeshcount; i++)
                {
                    if (MaterialValidForCollider(materials[i]))
                    {
                        foundValidMateiral = true;
                        break;
                    }
                }
                
                if (!foundValidMateiral)
                {// Nothing to do here...
                    return false;
                }
                
                Quaternion OldRot = GO.transform.rotation;
                Vector3 OldPos = GO.transform.position;

                GO.transform.rotation = Quaternion.identity;
                GO.transform.position = Vector3.zero;

                var originalMeshName = sharedMesh.name;

                // Make the new mesh...
                var newMesh = new Mesh();
                //newMesh.name = string.Format("GeneratedCollider_{0}", originalMeshName);
                newMesh.name = "GeneratedCollider_" + materials[0].name;//originalMeshName;
                newMesh.indexFormat = UnityEngine.Rendering.IndexFormat.UInt32;
                
                List<int> newTris = new List<int>();
                List<Vector3> newVerts = new List<Vector3>();
                List<Vector3> newNormals = new List<Vector3>();
                
                int current_vert = 0;

                for (int i = 0; i < totalsubmeshcount; i++)
                {
                    if (!MaterialValidForCollider(materials[i]))
                    {// Skip this one...
                        continue;
                    }
                    
                    var oldVerts = new List<Vector3>();
                    sharedMesh.GetVertices(oldVerts);

                    var oldTris = sharedMesh.GetTriangles(i);

                    var oldNormals = new List<Vector3>();
                    sharedMesh.GetNormals(oldNormals);
                    
                    bool haveNormals = true;
                    
                    if (oldNormals.Count <= 0)
                    {
                        haveNormals = false;
                    }
                    
                    var vertReplacements = new List<int>();
                    for (int j = 0; j < oldVerts.Count; j++)
                    {
                        vertReplacements.Add(-1);
                    }

                    for (int j = 0; j < oldTris.Length; j += 3)
                    {
                        Vector3 oldNormal1 = new Vector3(0,0,0);
                        Vector3 oldNormal2 = new Vector3(0,0,0);
                        Vector3 oldNormal3 = new Vector3(0,0,0);
                        
                        var oldTri1 = oldTris[j];
                        var oldVert1 = oldVerts[oldTri1];
                        
                        if (haveNormals) oldNormal1 = oldNormals[oldTri1];

                        var oldTri2 = oldTris[j + 1];
                        var oldVert2 = oldVerts[oldTri2];
                        if (haveNormals) oldNormal2 = oldNormals[oldTri2];

                        var oldTri3 = oldTris[j + 2];
                        var oldVert3 = oldVerts[oldTri3];
                        if (haveNormals) oldNormal3 = oldNormals[oldTri3];
                        
                        int found1 = vertReplacements[oldTri1];
                        int found2 = vertReplacements[oldTri2];
                        int found3 = vertReplacements[oldTri3];

                        if (heightRangeEnabledJSON.val && oldVert1.y < minHeightJSON.val && oldVert2.y < minHeightJSON.val && oldVert3.y < minHeightJSON.val)
                        {
                            continue;
                        }
                        
                        if (heightRangeEnabledJSON.val && oldVert1.y > maxHeightJSON.val && oldVert2.y > maxHeightJSON.val && oldVert3.y > maxHeightJSON.val)
                        {
                            continue;
                        }

                        if (found1 >= 0)
                        {// Reuse the old vert...
                            newTris.Add(found1);
                        }
                        else
                        {
                            newVerts.Add(oldVert1);
                            if (haveNormals) newNormals.Add(oldNormal1);
                            newTris.Add(current_vert);
                            current_vert++;
                            
                            vertReplacements[oldTri1] = newVerts.Count - 1;
                        }

                        if (found2 >= 0)
                        {// Reuse the old vert...
                            newTris.Add(found2);
                        }
                        else
                        {
                            newVerts.Add(oldVert2);
                            if (haveNormals) newNormals.Add(oldNormal2);
                            newTris.Add(current_vert);
                            current_vert++;
                            
                            vertReplacements[oldTri2] = newVerts.Count - 1;
                        }

                        if (found3 >= 0)
                        {// Reuse the old vert...
                            newTris.Add(found3);
                        }
                        else
                        {
                            newVerts.Add(oldVert3);
                            if (haveNormals) newNormals.Add(oldNormal3);
                            newTris.Add(current_vert);
                            current_vert++;
                            
                            vertReplacements[oldTri3] = newVerts.Count - 1;
                        }
                    }
                }
                
                if (current_vert <= 0 || newVerts.Count <= 0 || newTris.Count <= 0)
                {// Seems there was nothing to add (max height most likely)...
                    newMesh = null;
                    
                    GO.transform.rotation = OldRot;
                    GO.transform.position = OldPos;
                    return false;
                }
                else
                {
                    newMesh.SetVertices(newVerts);
                    newMesh.SetTriangles(newTris, 0);
                    newMesh.SetNormals(newNormals);
                    
                    var collider = GO.AddComponent<MeshCollider>();
                    collider.sharedMesh = null;
                    collider.sharedMesh = newMesh;
                    
                    newMesh.RecalculateBounds();
                    
                    bool useConvex = false;
                    
                    if (newMesh.bounds.size.x <= 0.1f && newMesh.bounds.size.y <= 0.1f && newMesh.bounds.size.z <= 0.1f)
                    {
                        //SuperController.LogMessage("Mesh " + newMesh.name + " is small (" + newMesh.bounds.size.x + " x " + newMesh.bounds.size.y + " x " + newMesh.bounds.size.z + "), will use a convex collider.");
                        useConvex = true;
                    }
                    else
                    {
                        //SuperController.LogMessage("Mesh " + newMesh.name + " is large (" + newMesh.bounds.size.x + " x " + newMesh.bounds.size.y + " x " + newMesh.bounds.size.z + "), will not use a convex collider.");
                    }
                    
                    if (convexCollidersAllowedJSON.val && useConvex)
                    {// See if we can make a convex collider...
                        try
                        {// Inside the try in case it fails, it should fall back...
                            collider.inflateMesh = true;
                            collider.skinWidth = 0.0001f;
                            collider.convex = true;
                            
                            //SuperController.LogMessage("GenerateCollider: Created collider mesh " + newMesh.name + " with " + newVerts.Count + " verts and " + newTris.Count + " tris.");
                            //SuperController.LogMessage("GenerateCollider: A convex collider was generated successfully from " + newMesh.name + ".");
                        }
                        catch (Exception e)
                        {
                            collider.inflateMesh = false;
                            collider.convex = false;
                            
                            //SuperController.LogMessage("GenerateCollider: Created collider mesh " + newMesh.name + " with " + newVerts.Count + " verts and " + newTris.Count + " tris.");
                            //SuperController.LogMessage("GenerateCollider: Convex collider failed to generate for this mesh, using full mesh for collisions.");
                        }
                    }
                    else
                    {
                        //SuperController.LogMessage("GenerateCollider: Created collider mesh " + newMesh.name + " with " + newVerts.Count + " verts and " + newTris.Count + " tris.");
                    }
                    
                    GO.transform.rotation = OldRot;
                    GO.transform.position = OldPos;
                }
            }
            catch (Exception e)
            {
                // Get stack trace for the exception with source file information
                var st = new StackTrace(e, true);
                // Get the top stack frame
                var frame = st.GetFrame(0);
                // Get the line number from the stack frame
                var line = frame.GetFileLineNumber();
                
                SuperController.LogError("Exception caught: line: " + line + "\n" + e);
                //SuperController.LogError("Exception caught: " + frame);
            }
            
            return true;
        }
        
        void DestroyColliders(GameObject sourceObject)
        {
            if (sourceObject != null)
            {
                var oldBoxColls = sourceObject.GetComponentsInChildren<BoxCollider>();
                
                foreach (BoxCollider collider in oldBoxColls)
                {
                    if (collider != null)
                    {
                        Destroy(collider);
                    }
                }
                
                var oldCapsuleColls = sourceObject.GetComponentsInChildren<CapsuleCollider>();
                
                foreach (CapsuleCollider collider in oldCapsuleColls)
                {
                    if (collider != null)
                    {
                        Destroy(collider);
                    }
                }
                
                var oldSphereColls = sourceObject.GetComponentsInChildren<SphereCollider>();
                
                foreach (SphereCollider collider in oldSphereColls)
                {
                    if (collider != null)
                    {
                        Destroy(collider);
                    }
                }
                
                var oldTerrainColls = sourceObject.GetComponentsInChildren<TerrainCollider>();
                
                foreach (TerrainCollider collider in oldTerrainColls)
                {
                    if (collider != null)
                    {
                        Destroy(collider);
                    }
                }
                
                var colliders = sourceObject.GetComponentsInChildren<MeshCollider>();
                
                if (colliders != null && colliders.Length > 0)
                {
                    foreach (MeshCollider collider in colliders)
                    {
                        if (collider != null)
                        {
                            //SuperController.LogMessage("deleting collider: " + collider.sharedMesh.name);
                            Destroy(collider);
                        }
                    }
                }
            }
        }
        
        //
        // Mesh Combiner/Optimizer...
        //
        
        class CombineData
        {
            public Material material;
            public List<MeshFilter> filter = new List<MeshFilter>();
            public List<MeshRenderer> renderer = new List<MeshRenderer>();
            public int lightmapIndex;
        }
        
        public bool is32bit = true;
        
        void CombineMeshes(GameObject combineParent)
        {
            try
            {
                // Verify there is existing object root, otherwise bail.
                if (combineParent == null)
                {
                    SuperController.LogMessage("OptimizeCUA: Parent of objects to combne not assigned. Operation cancelled.");
                    return;
                }

                // Remember the original position of the object. 
                // For the operation to work, the position must be temporarily set to (0,0,0).
                Vector3 originalPosition = combineParent.transform.position;
                combineParent.transform.position = Vector3.zero;

                // Locals
                List<CombineData> combinableMeshesList = new List<CombineData>();
                List<GameObject> combinedObjects = new List<GameObject>();
                
                //SuperController.LogMessage("DEBUG: Checking validity of meshes for combining...");

                MeshFilter[] meshFilters = combineParent.GetComponentsInChildren<MeshFilter>();
                
                // Go through all mesh filters and establish the mapping between the materials and all mesh filters using it.
                foreach (var meshFilter in meshFilters)
                {
                    var meshRenderer = meshFilter.GetComponent<MeshRenderer>();
                    
                    if (!meshFilter.sharedMesh || !meshFilter.sharedMesh.isReadable)
                    {
                        continue;
                    }
                    
                    if (meshRenderer == null)
                    {
                        SuperController.LogMessage("OptimizeCUA: The Mesh Filter on object " + meshFilter.name + " has no Mesh Renderer component attached. Skipping.");
                        continue;
                    }
                    
                    if (!meshRenderer.enabled)
                    {
                        continue;
                    }
                    
                    // First, clean up materials and change any material instances back to the base material...
                    if (meshRenderer.sharedMaterials.Length < 1 || meshRenderer.sharedMaterials[0] == null)
                    {
                        //combineParent.transform.position = originalPosition;
                        SuperController.LogMessage("OptimizeCUA: Material on mesh renderer " + meshRenderer.gameObject.name + " is null, skipping optimize on that renderer.");
                        continue;
                    }
                    
                    var materials = meshRenderer.sharedMaterials;
                    
                    if (materials == null)
                    {
                        //combineParent.transform.position = originalPosition;
                        SuperController.LogMessage("OptimizeCUA: The Mesh Renderer on object " + meshFilter.name + " has no material assigned. Skipping.");
                        continue;
                    }

                    // Continue...
                    var material = materials[0];
                    
                    //SuperController.LogMessage("MAT: " + material.name);

                    bool added = false;
                    
                    // Add material to mesh filter mapping to dictionary
                    foreach (var entry in combinableMeshesList)
                    {
                        if (entry.material == material)
                        {
                            if (entry.lightmapIndex == meshRenderer.lightmapIndex)
                            {// Only combine if the lightmap index matches, otherwise make a new copy of the material so lightmaps work right...
                                //SuperController.LogMessage("entry.lightmapIndex == meshRenderer.lightmapIndex");
                                CombineData newData2 = new CombineData();
                                entry.filter.Add(meshFilter);
                                entry.renderer.Add(meshRenderer);
                                added = true;
                                //SuperController.LogMessage("append " + material.name + " lm " + meshRenderer.lightmapIndex);
                                break;
                            }
                            /*else
                            {
                                SuperController.LogMessage("entry.lightmapIndex != meshRenderer.lightmapIndex");
                            }*/
                        }
                    }
                    
                    if (!added)
                    {
                        //SuperController.LogMessage("new entry " + material.name + " lm " + meshRenderer.lightmapIndex);
                        
                        CombineData newData = new CombineData();
                        newData.material = material;
                        newData.filter.Add(meshFilter);
                        newData.renderer.Add(meshRenderer);
                        newData.lightmapIndex = meshRenderer.lightmapIndex;
                        combinableMeshesList.Add(newData);
                    }
                }
                
                //SuperController.LogMessage("DEBUG: Merging materials...");

                // For each material, create a new merged object, in the scene and in the assets folder.
                foreach (var entry in combinableMeshesList)
                {
                    List<MeshFilter> filterList = entry.filter;
                    List<MeshRenderer> rendererList = entry.renderer;
                    
                    //SuperController.LogMessage("DEBUG: dah");
                    
                    if (entry.material == null)
                    {
                        SuperController.LogMessage("DEBUG: shit");
                    }
                    
                    // Create a convenient material name
                    string materialName = entry.material.ToString().Split(' ')[0];
                    
                    //SuperController.LogMessage("DEBUG: materialName: " + materialName);
                    
                    CombineInstance[] combine = new CombineInstance[filterList.Count];
                    
                    for (int i = 0; i < filterList.Count; i++)
                    {
                        //SuperController.LogMessage("DEBUG: begin filterList " + (int)(i+1) + " / " + filterList.Count);
                        
                        combine[i].mesh = filterList[i].sharedMesh;
                        combine[i].transform = filterList[i].transform.localToWorldMatrix;
                        combine[i].lightmapScaleOffset = rendererList[i].lightmapScaleOffset;
                        combine[i].realtimeLightmapScaleOffset = rendererList[i].realtimeLightmapScaleOffset;
                    }
                    
                    //SuperController.LogMessage("DEBUG: combine set up for " + materialName);

                    // Create a new mesh using the combined properties
                    var format = is32bit? IndexFormat.UInt32 : IndexFormat.UInt16;
                    
                    Mesh combinedMesh = new Mesh { indexFormat = format };
                    combinedMesh.CombineMeshes(combine, true, true, true); // UQ: Do lightmap too!!!
                    
                    //SuperController.LogMessage("DEBUG: combined " + materialName);

                    // Create asset
                    materialName = materialName + "_" + combinedMesh.GetInstanceID();
                    combinedMesh.name = "CombinedMeshes_" + materialName;
                    
                    //SuperController.LogMessage("DEBUG: new material name " + materialName);

                    // Create game object
                    //string goName = (combinableMeshesList.Count > 1) ? materialName : "CombinedMeshes_" + combineParent.name;
                    string goName = materialName;
                    GameObject combinedObject = new GameObject(goName);
                    var filter = combinedObject.AddComponent<MeshFilter>();
                    filter.sharedMesh = combinedMesh;
                    var renderer = combinedObject.AddComponent<MeshRenderer>();
                    renderer.sharedMaterial = entry.material;
                    combinedObjects.Add(combinedObject);
                    renderer.lightmapIndex = entry.lightmapIndex;
                    
                    //SuperController.LogMessage("DEBUG: finished combine for " + materialName);
                }
                
                //SuperController.LogMessage("DEBUG: Clearing old gameobject...");

                GameObject resultGO = new GameObject("CombinedMeshes_" + combineParent.name);
                
                // Rebuild the gameobject...
                foreach (var combinedObject in combinedObjects)
                {
                    combinedObject.transform.parent = resultGO.transform;
                }
                
                // Return to original positions
                combineParent.transform.position = originalPosition;
                resultGO.transform.position = originalPosition;
                //resultGO.transform.position = Vector3.zero;
                resultGO.transform.parent = combineParent.transform;
                
                foreach (var entry in combinableMeshesList)
                {
                    List<MeshRenderer> rendererList = entry.renderer;
                    
                    foreach (MeshRenderer r in rendererList)
                    {
                        r.enabled = false;
                    }
                }
            }
            catch (System.ArgumentException e) 
            {
                SuperController.LogError("ArgumentException: " + e);
            }
            catch (IndexOutOfRangeException e)
            {
                SuperController.LogError("IndexOutOfRangeException: " + e);
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        
        //
        // Mesh splitting...
        //
        
        static GameObject baseGO = null;
        
        void SplitMeshesIfRequired(GameObject combineParent)
        {
            // Verify there is existing object root, otherwise bail.
            if (combineParent == null)
            {
                SuperController.LogMessage("OptimizeCUA: Parent of objects to combne not assigned. Operation cancelled.");
                return;
            }
            
            baseGO = combineParent;

            List<Transform> transforms = new List<Transform>();
            MeshFilter[] meshFilters = combineParent.GetComponentsInChildren<MeshFilter>();
            
            // Go through all mesh filters and establish the mapping between the materials and all mesh filters using it.
            foreach (var meshFilter in meshFilters)
            {
                if (!meshFilter.sharedMesh || !meshFilter.sharedMesh.isReadable)
                {
                    continue;
                }
                
                var meshRenderer = meshFilter.GetComponent<MeshRenderer>();
                
                if (meshRenderer == null)
                {
                    //SuperController.LogMessage("OptimizeCUA: The Mesh Filter on object " + meshFilter.name + " has no Mesh Renderer component attached. Skipping.");
                    continue;
                }
                
                if (!meshRenderer.enabled)
                {
                    continue;
                }
                
                var materials = meshRenderer.sharedMaterials;
                
                if (materials == null)
                {
                    SuperController.LogMessage("OptimizeCUA: The Mesh Renderer on object " + meshFilter.name + " has no material assigned. Skipping.");
                    continue;
                }
                
                if (meshFilter.transform == null || meshFilter.transform.Equals(null))
                {
                    SuperController.LogMessage("OptimizeCUA: " + meshFilter.name + "'s game object has a null transform.");
                }
                
                if (materials.Length > 1)
                {
                    transforms.Add(meshFilter.transform);
                }
            }
            
            ApplyTransform(transforms.ToArray());
        }
        
        // Apply individual transforms for multiple selected transforms.
        // It will apply the top-level transforms first, then update
        // children, but not siblings if those aren't selected.
        public static void ApplyTransform(Transform[] transforms)
        {
            bool[] applied = new bool[transforms.Length];
            for (int i = 0; i < applied.Length; ++i) applied[i] = false;

            int applyCount = 0;

            while (applyCount != transforms.Length)
            {
                // Pass over list, finding unapplied transforms with no parents or applied parents.
                for (int i = 0; i < transforms.Length; ++i)
                {
                    // If this entry is unapplied...
                    if (!applied[i])
                    {
                        bool canApply = true; // assume we can apply the transform.

                        // Is the entry a child of an unapplied parent?
                        for (int j = 0; j < transforms.Length; ++j)
                        {
                            if (i == j) continue; // ignore same entry.

                            // If it's a child of unapplied parent,
                            // we can't apply the transform to [i] in this pass.
                            if (transforms[i].IsChildOf(transforms[j]) && !applied[j])
                            {
                                canApply = false;
                                break;
                            }
                        }

                        if (canApply)
                        {
                            ApplyTransform(transforms[i]);
                            applied[i] = true;
                            //SuperController.LogMessage("OptimizeCUA: Applied transform to " + transforms[i].name);
                            ++applyCount;
                        }
                    }
                }
            }
        }

        // Apply an individual transform.
        public static void ApplyTransform(Transform transform)
        {
            if (transform == null)
            {
                return;
            }
            
            try
            {
                var meshFilter = transform.GetComponent<MeshFilter>();
                var meshRenderer = transform.GetComponent<MeshRenderer>();
                
                if (meshFilter == null || meshRenderer == null)
                {
                    return;
                }
                
                if (!meshRenderer.enabled)
                {
                    return;
                }
                
                if (meshFilter.sharedMesh == null)
                {
                    return;
                }
                
                if (!meshFilter.sharedMesh.isReadable)
                {
                    return;
                }
                
                //SuperController.LogMessage("MeshSplit:: Splitting mesh for object (" + transform.name + ").");
                string originalMeshName = meshFilter.sharedMesh.name;
                
                // move required submeshes lower
                var totalsubmeshcount = meshRenderer.sharedMaterials.Length;

                List<Vector2> uv1s = new List<Vector2>();
                List<Vector2> uv2s = new List<Vector2>();
                List<Vector2> uv3s = new List<Vector2>();
                List<Vector2> uv4s = new List<Vector2>();
                meshFilter.sharedMesh.GetUVs(0, uv1s);
                meshFilter.sharedMesh.GetUVs(1, uv2s);
                meshFilter.sharedMesh.GetUVs(2, uv3s);
                meshFilter.sharedMesh.GetUVs(3, uv4s);

                var materials = meshRenderer.sharedMaterials;

                for (int i = 0; i < totalsubmeshcount; i++)
                {
                    var newMesh = new Mesh();//Instantiate(meshFilter.sharedMesh);
                    newMesh.indexFormat = UnityEngine.Rendering.IndexFormat.UInt32;
                    
                    //SuperController.LogMessage("MeshSplit:: Splitting mesh for object (" + transform.name + ") - subMesh " + (int)(i+1) + "/" + totalsubmeshcount);

                    string materialName = materials[i].name;
                    string new_mesh_name = "SplitMesh_" + materialName + "_" + (int)Mathf.Abs(newMesh.GetHashCode());
                    newMesh.name = new_mesh_name;
                    
                    //SuperController.LogMessage("DEBUG: hash");

                    var oldVerts = new List<Vector3>();
                    meshFilter.sharedMesh.GetVertices(oldVerts);
                    
                    var vertReplacements = new List<int>();
                    for (int j = 0; j < oldVerts.Count; j++)
                    {
                        vertReplacements.Add(-1);
                    }
                    
                    //SuperController.LogMessage("DEBUG: verts");

                    var oldTris = meshFilter.sharedMesh.GetTriangles(i);

                    var oldNormals = new List<Vector3>();
                    meshFilter.sharedMesh.GetNormals(oldNormals);
                    
                    bool haveNormals = true;
                    
                    if (oldNormals.Count <= 0)
                    {
                        haveNormals = false;
                    }

                    List<int> newTris = new List<int>();
                    List<Vector3> newVerts = new List<Vector3>();

                    List<Vector3> newNormals = new List<Vector3>();

                    List<Vector2> newUv1s = new List<Vector2>();
                    List<Vector2> newUv2s = new List<Vector2>();
                    List<Vector2> newUv3s = new List<Vector2>();
                    List<Vector2> newUv4s = new List<Vector2>();
                    
                    //SuperController.LogMessage("DEBUG: begin");

                    for (int j = 0; j < oldTris.Length; j++)
                    {
                        var oldTri = oldTris[j];
                        var oldVert = oldVerts[oldTri];
                        var oldNormal = new Vector3(0,0,0);
                        
                        if (haveNormals)
                        {
                            oldNormal = oldNormals[oldTri];
                        }

                        // First check if we already have a vert at the same position in the new mesh, re-use it if we do...
                        int found = vertReplacements[oldTri];

                        if (found >= 0)
                        {// Reuse the old vert...
                            newTris.Add(found);
                        }
                        else
                        {// Vert position not already in the list, add a new one...
                            newVerts.Add(oldVert);
                            
                            if (haveNormals) newNormals.Add(oldNormal);
                            
                            newTris.Add(newVerts.Count - 1);

                            if (uv1s.Count > 0 && uv1s.Count >= oldTri)
                            {
                                newUv1s.Add(uv1s[oldTri]);
                            }

                            if (uv2s.Count > 0 && uv2s.Count >= oldTri)
                            {
                                newUv2s.Add(uv2s[oldTri]);
                            }

                            if (uv3s.Count > 0 && uv3s.Count >= oldTri)
                            {
                                newUv3s.Add(uv3s[oldTri]);
                            }

                            if (uv4s.Count > 0 && uv4s.Count >= oldTri)
                            {
                                newUv4s.Add(uv4s[oldTri]);
                            }
                            
                            vertReplacements[oldTri] = newVerts.Count - 1;
                        }
                    }
                    
                    //SuperController.LogMessage("DEBUG: reged");

                    newMesh.SetVertices(newVerts);
                    newMesh.SetNormals(newNormals);
                    newMesh.SetTriangles(newTris, 0);
                    newMesh.SetUVs(0, newUv1s);
                    newMesh.SetUVs(1, newUv2s);
                    newMesh.SetUVs(2, newUv3s);
                    newMesh.SetUVs(3, newUv4s);
                    
                    //SuperController.LogMessage("DEBUG: newMesh");

                    var newGameObject = new GameObject();
                    //SuperController.LogMessage("DEBUG: createGO");

                    //newGameObject.transform.parent = combineParent.transform;
                    //newGameObject.transform.parent = baseGO.transform;
                    newGameObject.transform.parent = transform;
                    //newGameObject.transform.parent = meshFilter.transform.gameObject.transform;
                    //SuperController.LogMessage("DEBUG: transform");
                    
                    var newMeshFilter = newGameObject.AddComponent<MeshFilter>();
                    var newMeshRenderer = newGameObject.AddComponent<MeshRenderer>();
                    //SuperController.LogMessage("DEBUG: create R+F");
                    
                    
                    newGameObject.name = new_mesh_name;
                    //SuperController.LogMessage("DEBUG: named");
        

                    Vector3 zero;
                    zero.x = 0.0f;
                    zero.y = 0.0f;
                    zero.z = 0.0f;

                    Vector3 scale;
                    scale.x = 1.0f;
                    scale.y = 1.0f;
                    scale.z = 1.0f;
                    newGameObject.transform.localPosition = zero;
                    newGameObject.transform.localRotation = Quaternion.identity;
                    newGameObject.transform.localScale = scale;

/*
                    newGameObject.transform.position = transform.position;
                    newGameObject.transform.rotation = transform.rotation;
                    //newGameObject.transform.scale = transform.scale;
*/
                    Material[] newMats = new Material[1];
                    newMats.SetValue(materials[i], 0);
                    newMeshRenderer.sharedMaterials = newMats;
                    newMeshFilter.sharedMesh = newMesh;
                    
                    //SuperController.LogMessage("DEBUG: newGOset");
                    
                    
                    newMeshRenderer.lightmapScaleOffset = meshRenderer.lightmapScaleOffset;
                    newMeshRenderer.realtimeLightmapScaleOffset = meshRenderer.realtimeLightmapScaleOffset;
                    newMeshRenderer.lightmapIndex = meshRenderer.lightmapIndex;
                    
                    //SuperController.LogMessage("DEBUG: lightmap");
                }
                
                //SuperController.LogMessage("MeshSplit:: Completed");
                
                meshRenderer.enabled = false;
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        
        //
        // Check if we have access to the asset's meshes...
        //
        
        bool CheckMeshAccessability(GameObject combineParent)
        {
            // Verify there is existing object root, otherwise bail.
            if (combineParent == null)
            {
                SuperController.LogMessage("OptimizeCUA: Parent of objects to combne not assigned. Operation cancelled.");
                return false;
            }
                
            MeshFilter[] meshFilters = combineParent.GetComponentsInChildren<MeshFilter>();
            
            // Go through all mesh filters and establish the mapping between the materials and all mesh filters using it.
            foreach (var meshFilter in meshFilters)
            {
                if (!meshFilter.sharedMesh || !meshFilter.sharedMesh.isReadable)
                {
                    return false;
                }
            }
            
            return true;
        }
        
        //
        // Cleanup...
        //
        
        void CleanupDisabledRenderers(GameObject combineParent)
        {
            MeshRenderer[] meshRenderers = combineParent.GetComponentsInChildren<MeshRenderer>();

            foreach (MeshRenderer r in meshRenderers)
            {
                if (!r.enabled)
                {
                    MeshFilter mf = r.gameObject.GetComponent<MeshFilter>();
                    DestroyImmediate(r, false);
                    
                    if (mf != null)
                        DestroyImmediate(mf, false);
                }
            }
        }
        
        void CleanupEmptyGameObjects(GameObject combineParent)
        {
            int numDestroyed = 0;
            
            Transform[] transforms = combineParent.GetComponentsInChildren<Transform>();

            foreach (Transform t in transforms)
            {
                GameObject GO = t.gameObject;
                
                Component[] components = GO.GetComponents(typeof(Component));
                
                if (components.Length <= 1 && t.childCount <= 0)
                {
                    DestroyImmediate(GO, false);
                }
            }
            
            if (numDestroyed > 0)
            {
                SuperController.LogMessage("OptimizeCUA: Destroyed " + numDestroyed + " empty game objects.");
            }
        }
        
        //
        //
        //
        
        void CheckRecalculateNormals(GameObject combineParent)
        {
            try
            {
                int numMeshesFixed = 0;
                
                MeshFilter[] meshFilters = combineParent.GetComponentsInChildren<MeshFilter>();

                foreach (MeshFilter mf in meshFilters)
                {
                    if (mf.sharedMesh != null)
                    {// Check if it has normals, if not, build them now...
                        var oldNormals = new List<Vector3>();
                        mf.sharedMesh.GetNormals(oldNormals);
                        
                        bool haveNormals = true;
                        
                        if (oldNormals.Count <= 0)
                        {
                            haveNormals = false;
                        }
                        
                        if (!haveNormals)
                        {
                            mf.sharedMesh.RecalculateNormals();
                            mf.sharedMesh.RecalculateTangents();
                            numMeshesFixed++;
                        }
                    }
                }
                
                if (numMeshesFixed > 0)
                {
                    SuperController.LogMessage("OptimizeCUA: Generated missing mesh normals and tangents for " + numMeshesFixed + " meshes.");
                }
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        
        //
        // Fix missing materials...
        //
        
        void FixMissingMaterials(GameObject combineParent)
        {
            // Verify there is existing object root, otherwise bail.
            if (combineParent == null)
            {
                return;
            }
            
            List<Transform> transforms = new List<Transform>();
            MeshFilter[] meshFilters = combineParent.GetComponentsInChildren<MeshFilter>();
            
            // Go through all mesh filters and establish the mapping between the materials and all mesh filters using it.
            foreach (var meshFilter in meshFilters)
            {
                var meshRenderer = meshFilter.GetComponent<MeshRenderer>();
                
                if (meshRenderer == null)
                {
                    SuperController.LogMessage("OptimizeCUA: The Mesh Filter on object " + meshFilter.name + " has no Mesh Renderer component attached. Skipping.");
                    continue;
                }
                
                if (!meshRenderer.enabled)
                {
                    continue;
                }
                
                var mats = meshRenderer.sharedMaterials;
                
                if (mats == null || mats.Length < 1)
                {
                    Material[] newMats = new Material[1];
                        
                    if (meshRenderer.sharedMaterial != null)
                    {
                        newMats[0] = meshRenderer.sharedMaterial;
                        SuperController.LogMessage("Material on mesh filter " + meshFilter.gameObject.name + " was missing and is now " + newMats[0].name + ".");
                    }
                    else if (meshRenderer.material != null)
                    {
                        newMats[0] = meshRenderer.material;
                        SuperController.LogMessage("Material on mesh filter " + meshFilter.gameObject.name + " was missing and is now " + newMats[0].name + ".");
                    }
                    else
                    {
                        newMats[0] = new Material(Shader.Find("Standard"));
                        SuperController.LogMessage("Material on mesh filter " + meshFilter.gameObject.name + " was missing and is now " + newMats[0].name + ".");
                    }
                    
                    meshRenderer.sharedMaterials = newMats;
                    meshRenderer.sharedMaterial = null;
                    meshRenderer.material = null;
                }
                else
                {
                    Material[] newMats = new Material[mats.Length];
                
                    for (int k = 0; k < mats.Length; k++)
                    {
                        if (mats[k] == null)
                        {
                            if (meshRenderer.sharedMaterial != null)
                            {
                                newMats[k] = meshRenderer.sharedMaterial;
                                SuperController.LogMessage("Material " + k + " on mesh filter " + meshFilter.gameObject.name + " was missing and is now " + newMats[k].name + ".");
                            }
                            else if (meshRenderer.material != null)
                            {
                                newMats[k] = meshRenderer.material;
                                SuperController.LogMessage("Material " + k + " on mesh filter " + meshFilter.gameObject.name + " was missing and is now " + newMats[k].name + ".");
                            }
                            else
                            {
                                newMats[k] = new Material(Shader.Find("Standard"));
                                SuperController.LogMessage("Material " + k + " on mesh filter " + meshFilter.gameObject.name + " was missing and is now " + newMats[k].name + ".");
                            }
                        }
                        else
                        {
                            newMats[k] = mats[k];
                        }
                    }
                    
                    meshRenderer.sharedMaterials = newMats;
                }
            }
        }
        
        //
        // Scene load and enable/disable...
        //
        
        public void OnEnable()
        {
            StartCoroutine(WaitForAssetBundle());
        }

        private void OnSceneLoaded()
        {
            OnEnable();
        }
        
        void OnDestroy()
        {
        
        }
        
        public void OnDisable()
        {
            OnDestroy();
        }
    }
}
