﻿//#define DEBUG

using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using UnityEngine.Events;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;

namespace UQPlugin
{

    public class ScaleAnything : MVRScript
    {
        public string pluginAuthor = "unique";
        public string pluginName = "ScaleAnything";
        public string pluginVersion = "1.0";
        public string pluginDate = "2021-10-14";
        public string pluginDescription =
 @"
 Fine control over X/Y/Z scales of anything.
 ";
 
        public JSONStorableStringChooser precisionSelectionJSON;
        
        public JSONStorableFloat scaleXJSON;
        public JSONStorableFloat scaleYJSON;
        public JSONStorableFloat scaleZJSON;
        
        UIDynamicSlider scaleX;
        UIDynamicSlider scaleY;
        UIDynamicSlider scaleZ;
        
        GameObject scaleObject = null;
        
        float previousScaleX = 0.0f;
        float previousScaleY = 0.0f;
        float previousScaleZ = 0.0f;
        
        List<string> precisionsList = new List<string>();
        
        void CreatePrecisionsList()
        {
            precisionsList.Clear();
            precisionsList.Add("2");
            precisionsList.Add("3");
            precisionsList.Add("4");
            precisionsList.Add("5");
            precisionsList.Add("6");
            precisionsList.Add("7");
            precisionsList.Add("8");
            precisionsList.Add("9");
        }
        
        public void SetPrecision(string precision)
        {
            scaleX.valueFormat = "F" + precision;
            scaleY.valueFormat = "F" + precision;
            scaleZ.valueFormat = "F" + precision;
        }
        
        public override void Init() 
        {
            SetupScaleObject();
            
            JSONStorableString usageInfoTextJSON = new JSONStorableString("", "<size=10>\n</size><b><size=40>Scale Anything</size></b>\n\nGives fine control over X/Y/Z scales of anything. By default, VaM does not allow separate X/Y/Z scales for most things.\n\nNow you can scale any object any way you wish!");
            UIDynamicTextField usageInfoText = CreateTextField(usageInfoTextJSON);
            usageInfoText.backgroundColor = new Color32(60, 58, 210, 255);
            usageInfoText.height = 375f;
            usageInfoText.textColor = Color.white;
            
            //
            // Settings...
            //
            
            
            CreatePrecisionsList();
            
            JSONStorableString usageInfo2TextJSON = new JSONStorableString("", "\nNeed even more precision? Use this!");
            UIDynamicTextField usageInfo2 = CreateTextField(usageInfo2TextJSON, true);
            usageInfo2.backgroundColor = new Color32(210, 58, 60, 255);
            usageInfo2.height = 60f;
            usageInfo2.textColor = Color.white;
            
            
            precisionSelectionJSON = new JSONStorableStringChooser("Precision <size=18>(Decimal Points)</size>", precisionsList, "3", "Precision <size=18>(Decimal Points)</size>", SetPrecision);                
            RegisterStringChooser(precisionSelectionJSON);
            UIDynamicPopup dp2 = CreatePopup(precisionSelectionJSON, true);
            precisionSelectionJSON.storeType = JSONStorableParam.StoreType.Full;
            dp2.popupPanelHeight = 520f;
            dp2.popup.alwaysOpen = false;
            dp2.popup.onOpenPopupHandlers += CreatePrecisionsList;
            
            
            //
            // Positions...
            //
            
            scaleXJSON = new JSONStorableFloat("Scale X", 1.0f, (val) => { UpdateScale(val); }, -100.0f, 100.0f, false, true);
            scaleXJSON.storeType = JSONStorableParam.StoreType.Full;
            scaleX = CreateSlider(scaleXJSON, false);
            scaleX.valueFormat = "F3";
            RegisterFloat(scaleXJSON);
            
            scaleYJSON = new JSONStorableFloat("Scale Y", 1.0f, (val) => { UpdateScale(val); }, -100.0f, 100.0f, false, true);
            scaleYJSON.storeType = JSONStorableParam.StoreType.Full;
            scaleY = CreateSlider(scaleYJSON, false);
            scaleY.valueFormat = "F3";
            RegisterFloat(scaleYJSON);
            
            scaleZJSON = new JSONStorableFloat("Scale Z", 1.0f, (val) => { UpdateScale(val); }, -100.0f, 100.0f, false, true);
            scaleZJSON.storeType = JSONStorableParam.StoreType.Full;
            scaleZ = CreateSlider(scaleZJSON, false);
            scaleZ.valueFormat = "F3";
            RegisterFloat(scaleZJSON);
            
            SetPrecision(precisionSelectionJSON.val);
            
            scaleXJSON.val = 1.0f;
            scaleYJSON.val = 1.0f;
            scaleZJSON.val = 1.0f;
        }
        
        void SetupScaleObject()
        {
            if (scaleObject == null)
            {
                var sourceObject = containingAtom.containingSubScene == null ? containingAtom.gameObject : containingAtom.parentAtom.gameObject;
                var tansforms = sourceObject.GetComponentsInChildren<Transform>();
                
                foreach (Transform trans in tansforms)
                {
                    if (trans.gameObject.name == "rescaleObject")
                    {
                        scaleObject = trans.gameObject;
                        break;
                    }
                }
                
                if (scaleObject == null)
                {
                    scaleObject = sourceObject;
                }
                
                if (scaleObject == null)
                {
                    SuperController.LogMessage("ScaleAnything: Did not find a gameObject. DISABLED!");
                    return;
                }
                
                if (scaleObject.name != "rescaleObject")
                {
                    SuperController.LogMessage("ScaleAnything: Did not find a rescaleObject. You might see strange effects like UI scaling with the object when rescaled.");
                }
                
                //SuperController.LogMessage("DEBUG: parent Game object is " + scaleObject.name);
            }
        }
        
        void UpdateScale(float unused)
        {
            SetupScaleObject();
            
            if (scaleObject == null)
            {
                return;
            }
            
            Vector3 scale = new Vector3(scaleXJSON.val, scaleYJSON.val, scaleZJSON.val);
            scaleObject.transform.localScale = scale;
            
            previousScaleX = scaleXJSON.val;
            previousScaleY = scaleYJSON.val;
            previousScaleZ = scaleZJSON.val;
        }
        
        public void Update()
        {
            if (previousScaleX != scaleXJSON.val || previousScaleY != scaleYJSON.val || previousScaleZ != scaleZJSON.val)
            {
                UpdateScale(0.0f);
            }
        }
        
        //
        // Support for turning on/off and scene loading...
        //
        public void OnEnable()
        {
            UpdateScale(0.0f);
        }

        public void OnDisable()
        {
            // Meh, leave the changes.
            // Setting to zero is not good, recording pos/rot on load may not match external changes made after loading the plugin...
            // There's is just no real way to do this without causing other issues...
        }
        
        private void OnDestroy()
        {
            OnDisable();
        }
        
        private void OnSceneLoaded()
        {
            OnEnable();
        }
    }
}
